# Copyright (C) 2018 Simon Biggs
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at

#     http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


"""Gamma filters that only calculate pass or fail for each evaluation point.

No interpolation is undergone with these at the moment. Interpolation is
planned for the future.
"""


import sys

from pymedphys._imports import numpy as np

from ..utilities import (
    calculate_pass_rate,
    convert_to_ravel_index,
    create_point_combination,
)


def gamma_filter_numpy(
    axes_reference,
    dose_reference,
    axes_evaluation,
    dose_evaluation,
    distance_mm_threshold,
    dose_threshold,
    lower_dose_cutoff=0,
    **_
):

    coord_diffs = [
        coord_ref[:, None] - coord_eval[None, :]
        for coord_ref, coord_eval in zip(axes_reference, axes_evaluation)
    ]

    all_in_vicinity = [
        np.where(np.abs(diff) < distance_mm_threshold) for diff in coord_diffs
    ]

    ref_coord_points = create_point_combination(
        [in_vicinity[0] for in_vicinity in all_in_vicinity]
    )

    eval_coord_points = create_point_combination(
        [in_vicinity[1] for in_vicinity in all_in_vicinity]
    )

    distances = np.sqrt(
        np.sum(
            [
                coord_diff[ref_points, eval_points] ** 2
                for ref_points, eval_points, coord_diff in zip(
                    ref_coord_points, eval_coord_points, coord_diffs
                )
            ],
            axis=0,
        )
    )

    within_distance_threshold = distances < distance_mm_threshold

    distances = distances[within_distance_threshold]
    ref_coord_points = ref_coord_points[:, within_distance_threshold]
    eval_coord_points = eval_coord_points[:, within_distance_threshold]

    dose_diff = (
        dose_evaluation[
            eval_coord_points[0, :], eval_coord_points[1, :], eval_coord_points[2, :]
        ]
        - dose_reference[
            ref_coord_points[0, :], ref_coord_points[1, :], ref_coord_points[2, :]
        ]
    )

    gamma = np.sqrt(
        (dose_diff / dose_threshold) ** 2 + (distances / distance_mm_threshold) ** 2
    )

    gamma_pass = gamma < 1

    eval_pass = eval_coord_points[:, gamma_pass]

    ravel_index = convert_to_ravel_index(eval_pass)
    gamma_pass_array = np.zeros_like(dose_evaluation).astype(np.bool)

    gamma_pass_array = np.ravel(gamma_pass_array)
    dose_above_cut_off = np.ravel(dose_evaluation) > lower_dose_cutoff

    gamma_pass_array[ravel_index] = True
    gamma_pass_percentage = np.mean(gamma_pass_array[dose_above_cut_off]) * 100

    return gamma_pass_percentage


def gamma_filter_brute_force(
    axes_reference,
    dose_reference,
    axes_evaluation,
    dose_evaluation,
    distance_mm_threshold,
    dose_threshold,
    lower_dose_cutoff=0,
    **_
):

    xx_ref, yy_ref, zz_ref = np.meshgrid(*axes_reference, indexing="ij")
    gamma_array = np.ones_like(dose_evaluation).astype(np.float) * np.nan

    mesh_index = np.meshgrid(
        *[np.arange(len(coord_eval)) for coord_eval in axes_evaluation]
    )

    eval_index = np.reshape(np.array(mesh_index), (3, -1))
    run_index = np.arange(np.shape(eval_index)[1])
    np.random.shuffle(run_index)

    sys.stdout.write("    ")

    for counter, point_index in enumerate(run_index):
        i, j, k = eval_index[:, point_index]
        eval_x = axes_evaluation[0][i]
        eval_y = axes_evaluation[1][j]
        eval_z = axes_evaluation[2][k]

        if dose_evaluation[i, j, k] < lower_dose_cutoff:
            continue

        distance = np.sqrt(
            (xx_ref - eval_x) ** 2 + (yy_ref - eval_y) ** 2 + (zz_ref - eval_z) ** 2
        )

        dose_diff = dose_evaluation[i, j, k] - dose_reference

        gamma = np.min(
            np.sqrt(
                (dose_diff / dose_threshold) ** 2
                + (distance / distance_mm_threshold) ** 2
            )
        )

        gamma_array[i, j, k] = gamma

        if counter // 30 == counter / 30:
            percent_pass = str(np.round(calculate_pass_rate(gamma_array), decimals=1))
            sys.stdout.write(
                "\rPercent Pass: {0}% | Percent Complete: {1:.2f}%".format(
                    percent_pass, counter / np.shape(eval_index)[1] * 100
                )
            )
            sys.stdout.flush()

    return calculate_pass_rate(gamma_array)



# Copyright (C) 2019 Cancer Care Associates

# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at

#     http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from pymedphys._imports import numpy as np

import pymedphys._base.delivery

from . import extract


def get_delivery_data_items(single_icom_stream):
    shrunk_stream, mu = extract.extract(single_icom_stream, "Delivery MU")
    shrunk_stream, gantry = extract.extract(shrunk_stream, "Gantry")
    shrunk_stream, collimator = extract.extract(shrunk_stream, "Collimator")

    shrunk_stream, mlc = extract.extract_coll(shrunk_stream, b"MLCX", 160)
    mlc = np.array(mlc)
    mlc = mlc.reshape((80, 2))
    mlc = np.fliplr(np.flipud(mlc * 10))
    mlc[:, 1] = -mlc[:, 1]
    mlc = np.round(mlc, 10)

    shrunk_stream, jaw = extract.extract_coll(shrunk_stream, b"ASYMY", 2)
    jaw = np.round(np.array(jaw) * 10, 10)
    jaw = np.flipud(jaw)

    return mu, gantry, collimator, mlc, jaw


def delivery_from_icom_stream(icom_stream):
    icom_stream_points = extract.get_data_points(icom_stream)
    delivery_raw = [
        get_delivery_data_items(single_icom_stream)
        for single_icom_stream in icom_stream_points
    ]

    mu = np.array([item[0] for item in delivery_raw])
    diff_mu = np.concatenate([[0], np.diff(mu)])
    diff_mu[diff_mu < 0] = 0
    mu = np.cumsum(diff_mu)

    gantry = np.array([item[1] for item in delivery_raw])
    collimator = np.array([item[2] for item in delivery_raw])
    mlc = np.array([item[3] for item in delivery_raw])
    jaw = np.array([item[4] for item in delivery_raw])

    return mu, gantry, collimator, mlc, jaw


class DeliveryIcom(
    pymedphys._base.delivery.DeliveryBase  # pylint: disable = protected-access
):
    @classmethod
    def from_icom(cls, icom_stream):
        return cls(  # pylint: disable = protected-access
            *delivery_from_icom_stream(icom_stream)
        )._filter_cps()



#############################START LICENSE##########################################
# Copyright (C) 2019 Pedro Martinez
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at

#     http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#############################END LICENSE##########################################


"""
###########################################################################################
#
#   Script name: Oncentra_DataExplr
#
#   Description: Tool for processing of Oncentra dose files
#
#   Example usage: python Oncentra_DataExplr.py "/file/"
#
#   Author: Pedro Martinez
#   pedro.enrique.83@gmail.com
#   5877000722
#   Date:2019-05-10
#
###########################################################################################
"""

import argparse
import os

import pydicom


def process_directory(directy):
    """Process directory."""
    filetype = ""
    desc = ""
    print(directy)
    for subdir, dirs, files in os.walk(directy):  # pylint: disable = unused-variable
        for file in sorted(files):
            if os.path.splitext(file)[1] == ".dcm":
                dataset = pydicom.dcmread(os.path.join(subdir, file))
                # print(str.split(os.path.splitext(file)[0]))
                if str.split(os.path.splitext(file)[0])[0] == "RP":
                    try:
                        filetype = "Plan"
                        desc = dataset[0x300A, 0x0004].value
                    except:  # pylint: disable = bare-except
                        print("no description found")
                elif str.split(os.path.splitext(file)[0])[0] == "RD":
                    try:
                        filetype = "Dose"
                        desc = dataset[0x3004, 0x0006].value
                    except:  # pylint: disable = bare-except
                        print("no description found")
                elif str.split(os.path.splitext(file)[0])[0] == "RS":
                    # print(dataset)
                    # exit(0)
                    filetype = "Structure"
                    try:
                        desc = dataset[0x3006, 0x0006].value
                    except:  # pylint: disable = bare-except
                        print("no description found")
                else:
                    continue

                print(
                    "Type=", filetype, "Description=", desc, "Filepath=", directy + file
                )


if __name__ == "__main__":
    parser = argparse.ArgumentParser()
    parser.add_argument("-d", "--directory", help="path to folder")
    args = parser.parse_args()

    if args.directory:
        dirname = args.directory
        process_directory(dirname)



# Copyright (C) 2018 Paul King

# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at

#     http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import numpy as np

from pymedphys._data import download
from pymedphys.experimental.fileformats import read_prs


def test_read_prs():
    data_zip_name = "profiler_test_data.zip"

    file_name = download.get_file_within_data_zip(data_zip_name, "test_varian_open.prs")
    assert np.allclose(read_prs(file_name).cax, 45.50562901780488)
    assert np.allclose(read_prs(file_name).x[0][1], 0.579460838649598)
    assert np.allclose(read_prs(file_name).y[0][1], 0.2910764234184594)

    file_name = download.get_file_within_data_zip(
        data_zip_name, "test_varian_wedge.prs"
    )
    assert np.allclose(read_prs(file_name).cax, 21.863167869662274)
    assert np.allclose(read_prs(file_name).x[0][1], 0.5626051581458927)
    assert np.allclose(read_prs(file_name).y[0][1], 0.260042064635505)

    file_name = download.get_file_within_data_zip(data_zip_name, "test_tomo_50mm.prs")
    assert np.allclose(read_prs(file_name).cax, 784.320114110518)
    assert np.allclose(read_prs(file_name).x[0][1], 563.4064789252321)
    assert np.allclose(read_prs(file_name).y[0][1], 1.8690221773721463)



# Copyright (C) 2020 Cancer Care Associates

# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at

#     http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from pymedphys._imports import keyring, pymssql
from pymedphys._imports import streamlit as st

from pymedphys._mosaiq import helpers as msq_helpers
from pymedphys._streamlit.apps.metersetmap import _config, _deliveries
from pymedphys._streamlit.utilities import misc as st_misc
from pymedphys._streamlit.utilities import mosaiq as st_mosaiq


@st.cache(hash_funcs={pymssql.Cursor: id})
def get_patient_fields(cursor, patient_id):
    return msq_helpers.get_patient_fields(cursor, patient_id)


@st.cache(hash_funcs={pymssql.Cursor: id})
def get_patient_name(cursor, patient_id):
    return msq_helpers.get_patient_name(cursor, patient_id)


def mosaiq_input_method(patient_id="", key_namespace="", site=None, **_):
    mosaiq_details = _config.get_mosaiq_details()

    mosaiq_site = st_misc.site_picker(
        "Mosaiq Site", default=site, key=f"{key_namespace}_mosaiq_site"
    )

    server = mosaiq_details[mosaiq_site]["server"]
    st.write(f"Mosaiq Hostname: `{server}`")

    sql_user = keyring.get_password("MosaiqSQL_username", server)
    st.write(f"Mosaiq SQL login being used: `{sql_user}`")

    patient_id = st.text_input(
        "Patient ID", patient_id, key=f"{key_namespace}_patient_id"
    )
    st.write(patient_id)

    cursor = st_mosaiq.get_mosaiq_cursor(server)

    if patient_id == "":
        return {}

    patient_name = get_patient_name(cursor, patient_id)

    st.write(f"Patient Name: `{patient_name}`")

    patient_fields = get_patient_fields(cursor, patient_id)

    st.write(
        """
        #### Mosaiq patient fields
        """
    )

    patient_fields = patient_fields[patient_fields["monitor_units"] != 0]
    st.write(patient_fields)

    field_ids = patient_fields["field_id"]
    field_ids = field_ids.values.tolist()

    selected_field_ids = st.multiselect(
        "Select Mosaiq field id(s)", field_ids, key=f"{key_namespace}_mosaiq_field_id"
    )

    cursor_and_field_ids = [(cursor, field_id) for field_id in selected_field_ids]
    deliveries = _deliveries.cached_deliveries_loading(
        cursor_and_field_ids, _deliveries.delivery_from_mosaiq
    )
    identifier = f"{mosaiq_site} Mosaiq ({', '.join([str(field_id) for field_id in selected_field_ids])})"

    return {
        "site": mosaiq_site,
        "patient_id": patient_id,
        "patient_name": patient_name,
        "data_paths": [],
        "identifier": identifier,
        "deliveries": deliveries,
    }



# Copyright (C) 2020 Simon Biggs

# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at

#     http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


import warnings

from pymedphys._imports import numpy as np
from pymedphys._imports import scipy

from . import profiles


def create_bb_attenuation_func(diameter, penumbra, max_attenuation):
    dx = diameter / 100
    radius = diameter / 2
    image_half_width = penumbra * 2 + radius

    x = np.arange(-image_half_width, image_half_width + dx, dx)
    xx, yy = np.meshgrid(x, x)

    with warnings.catch_warnings():
        warnings.simplefilter("ignore")
        z = np.sqrt(radius ** 2 - xx ** 2 - yy ** 2) / radius

    z[np.isnan(z)] = 0

    sig = profiles.scaled_penumbra_sig() * penumbra
    sig_pixel = sig / dx

    filtered = scipy.ndimage.gaussian_filter(z, sig_pixel)
    interp = scipy.interpolate.RegularGridInterpolator(
        (x, x), filtered, bounds_error=False, fill_value=None
    )

    def attenuation(x, y):
        return 1 - interp((x, y)) * max_attenuation

    return attenuation


def create_field_with_bb_func(
    field_centre,
    field_side_lengths,
    field_penumbra,
    field_rotation,
    bb_centre,
    bb_diameter,
    bb_max_attenuation,
):
    field = profiles.create_rectangular_field_function(
        field_centre, field_side_lengths, field_penumbra, field_rotation
    )

    bb_penumbra = field_penumbra / 3
    bb_attenuation_map = create_bb_attenuation_func(
        bb_diameter, bb_penumbra, bb_max_attenuation
    )

    def field_with_bb(x, y):
        return field(x, y) * bb_attenuation_map(x - bb_centre[0], y - bb_centre[1])

    return field_with_bb


def stripes_artefact_func(x, attenuation=0.05, period=1.6):
    sin_result = np.sin(x * 2 * np.pi / period)
    return 1 - (sin_result + 1) / 2 * attenuation


def create_striped_field_func(field, attenuation=0.05, period=1.6):
    def striped_field(x, y):
        return stripes_artefact_func(x, attenuation, period) * field(x, y)

    return striped_field


def create_saturated_field_func(field, level=0.9):
    def saturated_field(x, y):
        result = field(x, y)
        result[result > level] = level

        return result

    return saturated_field



# Copyright (C) 2019 Paul King

# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at

#     http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import os

import numpy as np

from pymedphys._experimental.paulking.narrow_png import read_narrow_png as from_image

DATA_DIRECTORY = os.path.abspath(os.path.join(os.path.dirname(__file__), "data"))


def test_read_narrow_png():
    vert_strip = os.path.join(DATA_DIRECTORY, "FilmCalib_EBT_vert_strip.png")
    horz_strip = os.path.join(DATA_DIRECTORY, "FilmCalib_EBT_horz_strip.png")
    assert np.allclose(from_image(vert_strip)[0][0], from_image(horz_strip)[0][0])



# Copyright (C) 2015,2020 Cancer Care Associates

# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at

#     http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import re
from glob import glob

from pymedphys._imports import numpy as np
from pymedphys._imports import pandas as pd
from pymedphys._imports import plt
from pymedphys._imports import streamlit as st

import pymedphys.electronfactors as electronfactors

# Old code warning, the below is Simon Biggs from 2015... be nice to him


def visual_circle_and_ellipse(insert_x, insert_y, width, length, circle_centre):
    t = np.linspace(0, 2 * np.pi)
    circle = {
        "x": width / 2 * np.sin(t) + circle_centre[0],
        "y": width / 2 * np.cos(t) + circle_centre[1],
    }

    (
        x_shift,
        y_shift,
        rotation_angle,
    ) = electronfactors.visual_alignment_of_equivalent_ellipse(
        insert_x, insert_y, width, length, None
    )

    rotation_matrix = np.array(
        [
            [np.cos(rotation_angle), -np.sin(rotation_angle)],
            [np.sin(rotation_angle), np.cos(rotation_angle)],
        ]
    )

    ellipse = np.array([length / 2 * np.sin(t), width / 2 * np.cos(t)]).T

    rotated_ellipse = ellipse @ rotation_matrix

    translated_ellipse = rotated_ellipse + np.array([y_shift, x_shift])
    ellipse = {"x": translated_ellipse[:, 1], "y": translated_ellipse[:, 0]}

    return circle, ellipse


def plot_insert(insert_x, insert_y, width, length, circle_centre):
    circle, ellipse = visual_circle_and_ellipse(
        insert_x, insert_y, width, length, circle_centre
    )

    plt.figure()
    plt.plot(insert_x, insert_y)
    plt.axis("equal")

    plt.plot(circle["x"], circle["y"])
    plt.title("Insert shape parameterisation")
    plt.xlabel("x (cm)")
    plt.ylabel("y (cm)")
    plt.grid(True)

    plt.plot(ellipse["x"], ellipse["y"])


def plot_model(width_data, length_data, factor_data):
    i, j, k = electronfactors.create_transformed_mesh(
        width_data, length_data, factor_data
    )
    model_width, model_length, model_factor = i, j, k

    vmin = np.nanmin(np.concatenate([model_factor.ravel(), factor_data.ravel()]))
    vmax = np.nanmax(np.concatenate([model_factor.ravel(), factor_data.ravel()]))

    plt.scatter(
        width_data,
        length_data,
        s=100,
        c=factor_data,
        cmap="viridis",
        vmin=vmin,
        vmax=vmax,
        zorder=2,
    )

    plt.colorbar()

    cs = plt.contour(model_width, model_length, model_factor, 20, vmin=vmin, vmax=vmax)

    plt.clabel(cs, cs.levels[::2], inline=True)

    plt.title("Insert model")
    plt.xlabel("width (cm)")
    plt.ylabel("length (cm)")


def main():
    st.write(
        """
        # Electron Insert Factors
        """
    )

    patient_id = st.text_input("Patient ID")

    if patient_id == "":
        st.stop()

    rccc_string_search_pattern = r"\\monacoda\FocalData\RCCC\1~Clinical\*~{}\plan\*\*tel.1".format(
        patient_id
    )
    rccc_filepath_list = glob(rccc_string_search_pattern)

    nbccc_string_search_pattern = r"\\tunnel-nbcc-monaco\FOCALDATA\NBCCC\1~Clinical\*~{}\plan\*\*tel.1".format(
        patient_id
    )
    nbccc_filepath_list = glob(nbccc_string_search_pattern)

    sash_string_search_pattern = r"\\tunnel-sash-monaco\Users\Public\Documents\CMS\FocalData\SASH\1~Clinical\*~{}\plan\*\*tel.1".format(
        patient_id
    )
    sash_filepath_list = glob(sash_string_search_pattern)

    filepath_list = np.concatenate(
        [rccc_filepath_list, nbccc_filepath_list, sash_filepath_list]
    )

    electronmodel_regex = r"RiverinaAgility - (\d+)MeV"
    applicator_regex = r"(\d+)X\d+"

    insert_data = dict()  # type: ignore

    for telfilepath in filepath_list:
        insert_data[telfilepath] = dict()

        with open(telfilepath, "r") as file:
            telfilecontents = np.array(file.read().splitlines())

        insert_data[telfilepath]["reference_index"] = []
        for i, item in enumerate(telfilecontents):
            if re.search(electronmodel_regex, item):
                insert_data[telfilepath]["reference_index"] += [i]

        insert_data[telfilepath]["applicators"] = [
            re.search(applicator_regex, telfilecontents[i + 12]).group(
                1
            )  # type: ignore
            for i in insert_data[telfilepath]["reference_index"]
        ]

        insert_data[telfilepath]["energies"] = [
            re.search(electronmodel_regex, telfilecontents[i]).group(1)  # type: ignore
            for i in insert_data[telfilepath]["reference_index"]
        ]

    for telfilepath in filepath_list:
        with open(telfilepath, "r") as file:
            telfilecontents = np.array(file.read().splitlines())

        insert_data[telfilepath]["x"] = []
        insert_data[telfilepath]["y"] = []

        for i, index in enumerate(insert_data[telfilepath]["reference_index"]):
            insert_initial_range = telfilecontents[
                index + 51 : :
            ]  # coords start 51 lines after electron model name
            insert_stop = np.where(insert_initial_range == "0")[0][
                0
            ]  # coords stop right before a line containing 0

            insert_coords_string = insert_initial_range[:insert_stop]
            insert_coords = np.fromstring(",".join(insert_coords_string), sep=",")
            insert_data[telfilepath]["x"].append(insert_coords[0::2] / 10)
            insert_data[telfilepath]["y"].append(insert_coords[1::2] / 10)

    for telfilepath in filepath_list:
        insert_data[telfilepath]["width"] = []
        insert_data[telfilepath]["length"] = []
        insert_data[telfilepath]["circle_centre"] = []
        insert_data[telfilepath]["P/A"] = []

        for i in range(len(insert_data[telfilepath]["reference_index"])):

            width, length, circle_centre = electronfactors.parameterise_insert(
                insert_data[telfilepath]["x"][i], insert_data[telfilepath]["y"][i]
            )

            insert_data[telfilepath]["width"].append(width)
            insert_data[telfilepath]["length"].append(length)
            insert_data[telfilepath]["circle_centre"].append(circle_centre)

            insert_data[telfilepath]["P/A"].append(
                electronfactors.convert2_ratio_perim_area(width, length)
            )

    data_filename = r"S:\Physics\RCCC Specific Files\Dosimetry\Elekta_EFacs\electron_factor_measured_data.csv"
    data = pd.read_csv(data_filename)

    width_data = data["Width (cm @ 100SSD)"]
    length_data = data["Length (cm @ 100SSD)"]
    factor_data = data["RCCC Inverse factor (dose open / dose cutout)"]

    p_on_a_data = electronfactors.convert2_ratio_perim_area(width_data, length_data)

    for telfilepath in filepath_list:
        insert_data[telfilepath]["model_factor"] = []

        for i in range(len(insert_data[telfilepath]["reference_index"])):
            applicator = float(insert_data[telfilepath]["applicators"][i])
            energy = float(insert_data[telfilepath]["energies"][i])
            ssd = 100

            reference = (
                (data["Energy (MeV)"] == energy)
                & (data["Applicator (cm)"] == applicator)
                & (data["SSD (cm)"] == ssd)
            )

            number_of_measurements = np.sum(reference)

            if number_of_measurements < 8:
                insert_data[telfilepath]["model_factor"].append(np.nan)
            else:
                insert_data[telfilepath]["model_factor"].append(
                    electronfactors.spline_model_with_deformability(
                        insert_data[telfilepath]["width"],
                        insert_data[telfilepath]["P/A"],
                        width_data[reference],
                        p_on_a_data[reference],
                        factor_data[reference],
                    )[0]
                )

    for telfilepath in filepath_list:
        st.write("---")
        st.write("Filepath: `{}`".format(telfilepath))

        for i in range(len(insert_data[telfilepath]["reference_index"])):
            applicator = float(insert_data[telfilepath]["applicators"][i])
            energy = float(insert_data[telfilepath]["energies"][i])
            ssd = 100

            st.write(
                "Applicator: `{} cm` | Energy: `{} MeV`".format(applicator, energy)
            )

            width = insert_data[telfilepath]["width"][i]
            length = insert_data[telfilepath]["length"][i]

            plt.figure()
            plot_insert(
                insert_data[telfilepath]["x"][i],
                insert_data[telfilepath]["y"][i],
                insert_data[telfilepath]["width"][i],
                insert_data[telfilepath]["length"][i],
                insert_data[telfilepath]["circle_centre"][i],
            )

            reference = (
                (data["Energy (MeV)"] == energy)
                & (data["Applicator (cm)"] == applicator)
                & (data["SSD (cm)"] == ssd)
            )

            number_of_measurements = np.sum(reference)

            plt.figure()
            if number_of_measurements < 8:
                plt.scatter(
                    width_data[reference],
                    length_data[reference],
                    s=100,
                    c=factor_data[reference],
                    cmap="viridis",
                    zorder=2,
                )
                plt.colorbar()
            else:
                plot_model(
                    width_data[reference],
                    length_data[reference],
                    factor_data[reference],
                )

            reference_data_table = pd.concat(
                [width_data[reference], length_data[reference], factor_data[reference]],
                axis=1,
            )
            reference_data_table.sort_values(
                ["RCCC Inverse factor (dose open / dose cutout)"],
                ascending=False,
                inplace=True,
            )

            st.write(reference_data_table)

            st.pyplot()

            factor = insert_data[telfilepath]["model_factor"][i]

            st.write(
                "Width: `{0:0.2f} cm` | Length: `{1:0.2f} cm` | Factor: `{2:0.3f}`".format(
                    width, length, factor
                )
            )



# Copyright (C) 2020 Rafael Ayala

# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at

#     http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


import codecs
import datetime
import re
import socket

from pymedphys._imports import numpy as np
from pymedphys._imports import pandas as pd
from pymedphys._imports import tqdm


class QuickCheck:
    def __init__(self, ip):
        self.ip = ip
        self.port = 8123
        self.sock = socket.socket(socket.AF_INET, socket.SOCK_DGRAM)  # UDP
        self.sock.settimeout(3)
        self.MSG = b""
        self.raw_MSG = ""
        self.measurements = pd.DataFrame()
        self.data = ""
        self.raw_data = ""
        self.connected = False

    def connect(self):
        print("UDP target IP:", self.ip)
        print("UDP target port:", self.port)
        self.send_quickcheck("SER")
        if "SER" in self.data:
            self.connected = True
            print("Connected to Quickcheck")
            print(self.data)

    def close(self):
        self.sock.close()
        del self.sock

    def prepare_qcheck(self):
        self.MSG = (
            self.raw_MSG.encode()
            + codecs.decode("0d", "hex")
            + codecs.decode("0a", "hex")
        )

    def send_quickcheck(self, message):
        self.raw_MSG = message
        self.prepare_qcheck()
        max_retries = 3
        n_retry = 0

        while True:
            try:
                self.sock.sendto(self.MSG, (self.ip, self.port))
                self.raw_data, _ = self.sock.recvfrom(4096)
                data = self.raw_data.decode(encoding="utf-8")
                self.data = data.strip("\r\n")
                break
            except socket.timeout:
                if n_retry == max_retries:
                    print(
                        """
                          Connection Error  - Reached max retries
                          Quickcheck device unreachable, please check your settings"""
                    )
                    self.data = ""
                    break

                print("Connection Timeout")
                n_retry += 1
                print("Retrying connection {}/{}".format(n_retry, max_retries))

    def parse_measurements(self):
        data_split = self.data.split(";")
        m = {}  # Dictionary with measurements
        if data_split[0] == "MEASGET":
            #  MD section:__________________________________________________________
            MD = re.findall(r"MD=\[(.*?)\]", self.data)[0]
            m["MD_ID"] = np.int(re.findall(r"ID=(.*?);", MD)[0])
            meas_date = re.findall(r"Date=(.*?);", MD)[0]
            m["MD_Date"] = datetime.datetime.strptime(meas_date, "%Y-%m-%d").date()
            meas_time = re.findall(r"Time=(.*?)$", MD)[0]
            m["MD_Time"] = datetime.datetime.strptime(meas_time, "%H:%M:%S").time()
            m["MD_DateTime"] = datetime.datetime.combine(m["MD_Date"], m["MD_Time"])
            #  MV section:__________________________________________________________
            str_val = re.findall(r"MV=\[(.*?)\]", self.data)[0]
            regex_map = {
                "MV_CAX": "CAX=(.*?);",
                "MV_G10": "G10=(.*?);",
                "MV_L10": "L10=(.*?);",
                "MV_T10": "T10=(.*?);",
                "MV_R10": "R10=(.*?);",
                "MV_G20": "G20=(.*?);",
                "MV_L20": "L20=(.*?);",
                "MV_T20": "T20=(.*?);",
                "MV_R20": "R20=(.*?);",
                "MV_E1": "E1=(.*?);",
                "MV_E2": "E2=(.*?);",
                "MV_E3": "E3=(.*?);",
                "MV_E4": "E4=(.*?);",
                "MV_Temp": "Temp=(.*?);",
                "MV_Press": "Press=(.*?);",
                "MV_CAXRate": "CAXRate=(.*?);",
                "MV_ExpTime": "ExpTime=(.*?)$",
            }
            for key, pattern in regex_map.items():
                m[key] = np.float(re.findall(pattern, str_val)[0])

            #  AV section:__________________________________________________________
            AV = re.findall(r"AV=\[(.*?)\]\]", self.data)[0]
            AV = AV + "]"  # add last character ]
            for s in ("CAX", "FLAT", "SYMGT", "SYMLR", "BQF", "We"):
                str_val = re.findall(s + r"=\[(.*?)\]", AV)[0]
                m["AV_" + s + "_Min"] = np.float(re.findall("Min=(.*?);", str_val)[0])
                m["AV_" + s + "_Max"] = np.float(re.findall("Max=(.*?);", str_val)[0])
                m["AV_" + s + "_Target"] = np.float(
                    re.findall("Target=(.*?);", str_val)[0]
                )
                m["AV_" + s + "_Norm"] = np.float(re.findall("Norm=(.*?);", str_val)[0])
                m["AV_" + s + "_Value"] = np.float(
                    re.findall("Value=(.*?);", str_val)[0]
                )
                m["AV_" + s + "_Valid"] = np.int(re.findall("Valid=(.*?)$", str_val)[0])

            #  WORK section:__________________________________________________________
            str_val = re.findall(r"WORK=\[(.*?)\]", self.data)[0]

            m["WORK_ID"] = np.int(re.findall("ID=(.*?);", str_val)[0])
            m["WORK_Name"] = re.findall("Name=(.*?)$", str_val)[0]

            #  TASK section:__________________________________________________________
            str_val = re.findall(r"TASK=\[(.*?)\];MV", self.data)[0]
            m["TASK_ID"] = np.int(re.findall(r"ID=(.*?);", str_val)[0])
            m["TASK_TUnit"] = re.findall(r"TUnit=(.*?);", str_val)[0]
            m["TASK_En"] = np.int(re.findall(r"En=(.*?);", str_val)[0])
            m["TASK_Mod"] = re.findall(r"Mod=(.*?);", str_val)[0]
            m["TASK_Fs"] = re.findall(r"Fs=(.*?);", str_val)[0]
            m["TASK_SSD"] = np.int(re.findall(r"SDD=(.*?);", str_val)[0])
            m["TASK_Ga"] = np.int(re.findall(r"Ga=(.*?);", str_val)[0])
            m["TASK_We"] = np.int(re.findall(r"We=(.*?);", str_val)[0])
            m["TASK_MU"] = np.int(re.findall(r"MU=(.*?);", str_val)[0])
            m["TASK_My"] = np.float(re.findall(r"My=(.*?);", str_val)[0])
            m["TASK_Info"] = re.findall(r"Info=(.*?)$", str_val)[0]

            str_val = re.findall(r"Prot=\[(.*?)\];", str_val)[0]
            m["TASK_Prot_Name"] = re.findall(r"Name=(.*?);", str_val)[0]
            m["TASK_Prot_Flat"] = np.int(re.findall(r"Flat=(.*?);", str_val)[0])
            m["TASK_Prot_Sym"] = np.int(re.findall(r"Sym=(.*?)$", str_val)[0])
        elif data_split[0] == "MEASCNT":
            m[data_split[0]] = np.int(data_split[1:][0])
        elif data_split[0] in ("PTW", "SER", "KEY"):
            m[data_split[0]] = data_split[1:]
        return m

    def get_measurements(self):
        self.send_quickcheck("MEASCNT")
        if "MEASCNT" not in self.data:
            self.send_quickcheck("MEASCNT")
        m = self.parse_measurements()
        n_meas = m["MEASCNT"]
        print("Receiving Quickcheck measurements")
        meas_list = []
        for m in tqdm.tqdm(range(n_meas)):
            control = False
            while not control:
                self.send_quickcheck("MEASGET;INDEX-MEAS=" + "%d" % (m,))
                control = self.raw_MSG in self.data

            meas = self.parse_measurements()
            meas_list.append(meas)
        self.measurements = pd.DataFrame(meas_list)



# Copyright (C) 2017 Cancer Care Associates

# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at

#     http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


import os

from pymedphys._imports import numpy as np
from pymedphys._imports import pandas as pd


def file_output(output_directory, distance, relative_dose, scan_curvetype, scan_depth):
    """Store the loaded mephysto data into csv files for easy user confirmation
    and use.
    """
    # Determines the filepaths for the output
    filepaths = determine_output_filepaths(output_directory, scan_curvetype, scan_depth)

    columns = ["distance (mm)", "relative dose"]

    # Loop over each curvetype and save the data to csv
    for i, _ in enumerate(scan_curvetype):
        # Stacks the data into one array and transposes into column orientation
        data = np.vstack([distance[i], relative_dose[i]]).T

        # Use pandas to save data to csv
        df = pd.DataFrame(data, columns=columns)
        df.to_csv(filepaths[i])


def determine_output_filepaths(output_directory, scan_curvetype, scan_depth):
    """Determine a useful filepath for the saving of each mephysto scan.
    """
    filepaths = []

    # Loop over each scan curvetype creating a relevant filepath
    for i, curvetype in enumerate(scan_curvetype):
        if curvetype == "PDD":
            # Create the filename to be pdd_[number].csv
            filepaths.append(
                os.path.join(output_directory, "pdd_[{0:d}].csv".format(i))
            )

        elif curvetype == "INPLANE_PROFILE":
            # Create the filename to be inplaneprofile_depth_[number].csv
            filepaths.append(
                os.path.join(
                    output_directory,
                    "inplaneprofile_{0:d}mm_[{1:d}].csv".format(int(scan_depth[i]), i),
                )
            )

        elif curvetype == "CROSSPLANE_PROFILE":
            # Create the filename to be crossplaneprofile_depth_[number].csv
            filepaths.append(
                os.path.join(
                    output_directory,
                    "crossplaneprofile_{0:d}mm_[{1:d}].csv".format(
                        int(scan_depth[i]), i
                    ),
                )
            )

        else:
            # Raise an error if the curve type was not as expected
            raise ValueError("Unexpected scan_curvetype")

    return filepaths



# Copyright (C) 2020 Stuart Swerdloff, Simon Biggs
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at

#     http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import functools
import json
import logging
import pathlib
from os.path import abspath, basename, dirname, isdir, isfile
from os.path import join as pjoin

from pymedphys._imports import pydicom

from pymedphys._dicom.anonymise import (
    anonymise_dataset,
    anonymise_directory,
    anonymise_file,
    get_baseline_keyword_vr_dict,
    get_default_identifying_keywords,
)

from . import strategy

HERE = dirname(abspath(__file__))

IDENTIFYING_UIDS_FILEPATH = pjoin(HERE, "identifying_uids.json")


@functools.lru_cache()
def _get_default_identifying_uids():
    with open(IDENTIFYING_UIDS_FILEPATH) as uid_file:
        IDENTIFYING_UIDS = json.load(uid_file)
    return tuple(IDENTIFYING_UIDS)


def get_default_identifying_uids():
    return list(_get_default_identifying_uids())


@functools.lru_cache()
def _get_default_pseudonymisation_keywords():
    anon_keyword_list = get_default_identifying_keywords()
    # The preferred approach is to pseudonymise the contents
    # of sequences, rather than operate on the sequence itself
    #
    # Eliminating the keywords that are sequences fixes issue #1034
    # for default usage
    identifying_keywords_less_sequences = [
        x for x in anon_keyword_list if not x.endswith("Sequence")
    ]
    anon_keyword_set = set(identifying_keywords_less_sequences)
    pseudo_uid_set = set(get_default_identifying_uids())
    return tuple(anon_keyword_set.union(pseudo_uid_set))


def get_default_pseudonymisation_keywords():
    return list(_get_default_pseudonymisation_keywords())


def anonymise_with_pseudo_cli(args):

    if args.delete_unknown_tags:
        handle_unknown_tags = True
    elif args.ignore_unknown_tags:
        handle_unknown_tags = False
    else:
        handle_unknown_tags = None

    if not args.keywords_to_leave_unchanged:
        keywords_to_leave_unchanged = ()
    else:
        keywords_to_leave_unchanged = args.keywords_to_leave_unchanged

    replacement_strategy = None
    if args.pseudo:
        logging.info("Was run with pseudo!")
        identifying_keywords_for_pseudo = get_default_pseudonymisation_keywords()
        logging.info("Using pseudonymisation keywords")
        replacement_strategy = strategy.pseudonymisation_dispatch
        logging.info("Using pseudonymisation strategy")

    if isfile(args.input_path):
        anonymise_file(
            dicom_filepath=args.input_path,
            output_filepath=args.output_path,
            delete_original_file=args.delete_original_files,
            anonymise_filename=not args.preserve_filenames,
            replace_values=not args.clear_values,
            keywords_to_leave_unchanged=keywords_to_leave_unchanged,
            delete_private_tags=not args.keep_private_tags,
            delete_unknown_tags=handle_unknown_tags,
            replacement_strategy=replacement_strategy,
            identifying_keywords=identifying_keywords_for_pseudo,
        )

    elif isdir(args.input_path):
        anonymise_directory(
            dicom_dirpath=args.input_path,
            output_dirpath=args.output_path,
            delete_original_files=args.delete_original_files,
            anonymise_filenames=not args.preserve_filenames,
            replace_values=not args.clear_values,
            keywords_to_leave_unchanged=keywords_to_leave_unchanged,
            delete_private_tags=not args.keep_private_tags,
            delete_unknown_tags=handle_unknown_tags,
            replacement_strategy=replacement_strategy,
            identifying_keywords=identifying_keywords_for_pseudo,
        )

    else:
        raise FileNotFoundError(
            "No file or directory was found at the supplied input path."
        )


def is_valid_strategy_for_keywords(
    identifying_keywords=None, replacement_strategy=None
):
    if identifying_keywords is None:
        identifying_keywords = get_default_pseudonymisation_keywords()

    if replacement_strategy is None:
        replacement_strategy = strategy.pseudonymisation_dispatch

    baseline_keyword_vr_dict = get_baseline_keyword_vr_dict()
    for keyword in identifying_keywords:
        vr = baseline_keyword_vr_dict[keyword]
        # pydicom.datadict.dictionary_VR(pydicom.datadict.tag_for_keyword(keyword))
        if vr not in replacement_strategy:
            return False
    return True


def pseudonymise(dicom_input, output_path=None):
    """Convenient API to pseudonymisation.
    Elements whose tags are not in the pydicom dictionary will be deleted
    PatientSex will not be modified/pseudonymised
    For fine tune control, use anonymise_dataset() instead

    Parameters
    ----------
    dicom_input : ``pydicom.dataset.Dataset | str | pathlib.Path``
        Either a dataset, a path to a file or a path to a directory
    output_path : ``str | pathlib.Path``, optional
        If the input is a file or a path, the directory to place the
        pseudonymised files, by default None

    Returns
    -------
    ``pydicom.dataset.Dataset`` | ``str`` | ``list`` of ``str``
        if the dicom_input was a dataset, return the pseudonymised dataset
        if the dicom input was a file, return the path to the pseudonymised file.
        if the dicom input was a directory, return the list of successfully
        anonymised files, and return that instead of None
    """
    replacement_strategy = strategy.pseudonymisation_dispatch
    identifying_keywords_for_pseudo = get_default_pseudonymisation_keywords()
    if not is_valid_strategy_for_keywords():
        logging.error("Pseudonymisation strategy is not valid for keywords")
        logging.error("Please submit issue to PyMedPhys")
        # but continue on, the data might not contain the offending keywords
        # and if it does... there will be some kind of error raised
    keywords_to_leave_unchanged = list("PatientSex")

    if isinstance(dicom_input, pydicom.dataset.Dataset):
        pseudo_ds = anonymise_dataset(
            dicom_input,
            keywords_to_leave_unchanged=keywords_to_leave_unchanged,
            delete_unknown_tags=True,
            replacement_strategy=replacement_strategy,
            identifying_keywords=identifying_keywords_for_pseudo,
        )
        return pseudo_ds

    if pathlib.Path().joinpath(dicom_input).is_dir():
        pseudonymised_file_list = anonymise_directory(
            dicom_input,
            output_dirpath=output_path,
            keywords_to_leave_unchanged=keywords_to_leave_unchanged,
            delete_unknown_tags=True,
            replacement_strategy=replacement_strategy,
            identifying_keywords=identifying_keywords_for_pseudo,
        )
        return pseudonymised_file_list

    if not pathlib.Path().joinpath(dicom_input).is_file():
        raise FileNotFoundError(f"Unable to find {dicom_input}")
    pseudonymised_filepath = anonymise_file(
        dicom_input,
        output_filepath=output_path,
        keywords_to_leave_unchanged=keywords_to_leave_unchanged,
        delete_unknown_tags=True,
        replacement_strategy=replacement_strategy,
        identifying_keywords=identifying_keywords_for_pseudo,
    )
    return pseudonymised_filepath



# Copyright (C) 2018 Paul King

# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at

#     http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import os

import numpy as np

import pymedphys
from pymedphys.experimental.fileformats import read_mapcheck

DATA_DIRECTORY = os.path.abspath(os.path.join(os.path.dirname(__file__), "data"))


def test_read_mapcheck_txt():
    """Test for successful read of mapcheck file with consistent results."""

    file_path = pymedphys.data_path("tomo_mapcheck_test.txt")

    result = read_mapcheck(file_path)
    assert result.dose.shape == (len(result.x), len(result.y))  # x,y -> (x,y)
    assert result.x[0] < result.x[-1]  # x ascending
    assert result.y[0] < result.y[-1]  # y ascending
    assert len(set(result.x)) == len(set(result.x))  # x vals unique
    assert len(set(result.y)) == len(set(result.y))  # y vals unique
    assert np.all([i >= 0 for i in result.dose.flatten()])  # doses >= 0
    assert np.allclose(result.dose[20][20], 92.03876716076289)



"""A suite of functions to determine the MetersetMap for a range of
formats

.. WARNING::
   Although this is a useful tool in the toolbox for patient specific
   IMRT QA, in and of itself it is not a sufficient replacement for
   direct dosimetric measurement. This tool does not verify that the
   reported dose within the treatment planning system matches that
   delivered by the Linac.

   Deficiencies in a plan's agreement between the treatment planning
   system and the Linac due to deviations such as small field
   output factors or the MLC model will not be able to be picked up by
   this tool.

   Furthermore, when the MetersetMap is calculated off of log files
   alone this uses leaf and jaw positions as reported by the Linac which
   may not agree with their position in reality due to issues such as a
   hardware or calibration fault.
"""

# pylint: disable = unused-import

import textwrap

from ._metersetmap.metersetmap import calc_metersetmap as calculate
from ._metersetmap.metersetmap import display_metersetmap as display
from ._metersetmap.metersetmap import get_grid as grid

WARNING_MESSAGE = textwrap.dedent(__doc__.split("WARNING::")[-1])



# Copyright (C) 2020 Simon Biggs, Matthew Cooper

# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at

#     http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import warnings

from pymedphys._imports import numpy as np
from pymedphys._imports import plt, skimage


def get_image_transformation_parameters(dcm_ct):
    position = dcm_ct.ImagePositionPatient
    spacing = dcm_ct.PixelSpacing
    orientation = dcm_ct.ImageOrientationPatient

    dx, dy = spacing
    Cx, Cy, *_ = position
    Ox, Oy = orientation[0], orientation[4]

    return dx, dy, Cx, Cy, Ox, Oy


def reduce_expanded_mask(expanded_mask, img_size, expansion):
    return np.mean(
        np.mean(
            np.reshape(expanded_mask, (img_size, expansion, img_size, expansion)),
            axis=1,
        ),
        axis=2,
    )


def get_grid(dcm_ct):
    dx, dy, Cx, Cy, Ox, Oy = get_image_transformation_parameters(dcm_ct)

    ct_size = np.shape(dcm_ct.pixel_array)
    x_grid = np.arange(Cx, Cx + ct_size[0] * dx * Ox, dx * Ox)
    y_grid = np.arange(Cy, Cy + ct_size[1] * dy * Oy, dy * Oy)

    return x_grid, y_grid, ct_size


def calculate_expanded_mask(contours, dcm_ct, expansion):
    dx, dy, Cx, Cy, Ox, Oy = get_image_transformation_parameters(dcm_ct)

    ct_size = np.shape(dcm_ct.pixel_array)

    new_ct_size = np.array(ct_size) * expansion

    expanded_mask = np.zeros(new_ct_size)

    for xyz in contours:
        x = np.array(xyz[0::3])
        y = np.array(xyz[1::3])
        z = xyz[2::3]

        if len(set(z)) != 1:
            raise ValueError("Expected only one z value for a given contour")

        r = (((y - Cy) / dy * Oy)) * expansion + (expansion - 1) * 0.5
        c = (((x - Cx) / dx * Ox)) * expansion + (expansion - 1) * 0.5

        expanded_mask = np.logical_or(
            expanded_mask,
            skimage.draw.polygon2mask(new_ct_size, np.array(list(zip(r, c)))),
        )

    return expanded_mask


def calculate_anti_aliased_mask(contours, dcm_ct, expansion=5):
    expanded_mask = calculate_expanded_mask(contours, dcm_ct, expansion)

    x_grid, y_grid, ct_size = get_grid(dcm_ct)

    mask = reduce_expanded_mask(expanded_mask, ct_size[0], expansion)
    mask = 2 * mask - 1

    return x_grid, y_grid, mask


def get_contours_from_mask(x_grid, y_grid, mask):
    with warnings.catch_warnings():
        warnings.simplefilter("ignore", UserWarning)
        fig, ax = plt.subplots()
        cs = ax.contour(x_grid, y_grid, mask, [0])

    contours = [path.vertices for path in cs.collections[0].get_paths()]
    plt.close(fig)

    return contours



# Copyright (C) 2019 Cancer Care Associates

# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at

#     http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


import functools
import json
import logging
import os
import pathlib
import urllib.error
import urllib.request
import zipfile

from pymedphys._imports import tqdm

import pymedphys._utilities.filehash
from pymedphys import _config as pmp_config

from . import retry, zenodo

HERE = pathlib.Path(__file__).resolve().parent
DEFAULT_HASHES_PATH = HERE.joinpath("hashes.json")


@functools.lru_cache()
def create_download_progress_bar():
    class DownloadProgressBar(tqdm.tqdm):
        def update_to(self, b=1, bsize=1, tsize=None):
            if tsize is not None:
                self.total = tsize
            self.update(b * bsize - self.n)

    return DownloadProgressBar


@retry.retry((urllib.error.HTTPError, ConnectionResetError))
def download_with_progress(url, filepath):
    DownloadProgressBar = create_download_progress_bar()

    with DownloadProgressBar(
        unit="B", unit_scale=True, miniters=1, desc=url.split("/")[-1]
    ) as t:
        urllib.request.urlretrieve(url, filepath, reporthook=t.update_to)


def get_data_dir():
    data_dir = pmp_config.get_config_dir().joinpath("data")
    data_dir.mkdir(exist_ok=True)

    return data_dir


def get_file_within_data_zip(zip_name, file_name):
    dose_data_files = pymedphys.zip_data_paths(zip_name)
    path_match = [path for path in dose_data_files if path.name == file_name]

    if len(path_match) != 1:
        print(path_match)

        raise ValueError("Expected to find exactly one file")

    return str(path_match[0])


@functools.lru_cache()
def get_url_map():
    with open(HERE.joinpath("urls.json"), "r") as f:
        url_map = json.load(f)

    return url_map


def get_url(filename):
    filename = str(filename).replace(os.sep, "/")
    url_map = get_url_map()

    try:
        url = url_map[filename]
    except KeyError:
        raise ValueError("The file provided isn't within pymedphys' urls.json record.")

    return url


def download_all():
    paths = []
    for file_name in get_url_map().keys():
        paths.append(data_path(file_name))

    return paths


def data_path(
    filename,
    check_hash=True,
    redownload_on_hash_mismatch=True,
    delete_when_no_hash_found=True,
    url=None,
    hash_filepath=None,
):
    filename = str(filename).replace(os.sep, "/")
    filepath = get_data_dir().joinpath(filename)

    containing_directory = pathlib.Path(filepath).parent
    containing_directory.mkdir(exist_ok=True, parents=True)

    logging.debug("Filepath saving to is %s", filepath)
    logging.debug("Does filepath exist? %s", filepath.exists())

    if check_hash and filepath.exists():
        try:
            get_cached_filehash(filename, hash_filepath=hash_filepath)
        except NoHashFound:
            if delete_when_no_hash_found:
                logging.warning("No hash found, deleting current file")
                filepath.unlink()  # Force a redownload

    if not filepath.exists():
        if url is None:
            url = get_url(filename)

        download_with_progress(url, filepath)

    if check_hash:
        try:
            hash_agrees = data_file_hash_check(filename, hash_filepath=hash_filepath)
        except NoHashFound:
            return filepath.resolve()

        if not hash_agrees:
            if redownload_on_hash_mismatch:
                filepath.unlink()
                return data_path(
                    filename,
                    redownload_on_hash_mismatch=False,
                    url=url,
                    hash_filepath=hash_filepath,
                )

            raise ValueError("The file on disk does not match the recorded hash.")

    return filepath.resolve()


class NoHashFound(KeyError):
    pass


def get_cached_filehash(filename, hash_filepath=None):
    if hash_filepath is None:
        hash_filepath = DEFAULT_HASHES_PATH

    filename = str(filename).replace(os.sep, "/")

    with open(hash_filepath, "r") as hash_file:
        hashes = json.load(hash_file)

    try:
        cached_filehash = hashes[filename]
    except KeyError:
        logging.warning("No hash found for file '%s'", filename)
        logging.debug("Hashes found were %s", hashes.keys())
        raise NoHashFound

    return cached_filehash


def data_file_hash_check(filename, hash_filepath=None):
    if hash_filepath is None:
        hash_filepath = DEFAULT_HASHES_PATH

    filename = str(filename).replace(os.sep, "/")

    filepath = get_data_dir().joinpath(filename)
    calculated_filehash = pymedphys._utilities.filehash.hash_file(  # pylint: disable = protected-access
        filepath
    )

    logging.debug("Calculated filehash is %s", calculated_filehash)

    try:
        cached_filehash = get_cached_filehash(filename, hash_filepath=hash_filepath)

        logging.debug("Cached filehash is %s", cached_filehash)
    except NoHashFound:
        logging.warning("Hash not found in %s. File will be updated.", hash_filepath)
        with open(hash_filepath, "r") as hash_file:
            hashes = json.load(hash_file)

        hashes[filename] = calculated_filehash

        with open(hash_filepath, "w") as hash_file:
            json.dump(hashes, hash_file, indent=2, sort_keys=True)

        raise

    return cached_filehash == calculated_filehash


def zenodo_data_paths(
    record_name, check_hash=True, redownload_on_hash_mismatch=True, filenames=None
):
    file_urls = zenodo.get_zenodo_file_urls(record_name)

    if filenames is not None:
        file_urls = {
            filename: url
            for filename, url in file_urls.items()
            if filename in filenames
        }

    logging.debug("File URLS are %s", file_urls)

    record_directory = get_data_dir().joinpath(record_name)
    record_directory.mkdir(exist_ok=True)

    relative_record_path = pathlib.Path(record_name)

    data_paths = []
    for filename, url in file_urls.items():
        filename = pathlib.Path(filename)
        save_filename = relative_record_path.joinpath(filename)

        if filename.suffix == ".zip":
            data_paths += zip_data_paths(
                save_filename,
                check_hash=check_hash,
                redownload_on_hash_mismatch=redownload_on_hash_mismatch,
                url=url,
            )
        else:
            data_paths.append(
                data_path(
                    save_filename,
                    check_hash=check_hash,
                    redownload_on_hash_mismatch=redownload_on_hash_mismatch,
                    url=url,
                )
            )

    return data_paths


def zip_data_paths(
    filename,
    check_hash=True,
    redownload_on_hash_mismatch=True,
    delete_when_no_hash_found=True,
    url=None,
    extract_directory=None,
    hash_filepath=None,
):
    zip_filepath = data_path(
        filename,
        check_hash=check_hash,
        redownload_on_hash_mismatch=redownload_on_hash_mismatch,
        delete_when_no_hash_found=delete_when_no_hash_found,
        url=url,
        hash_filepath=hash_filepath,
    )

    if extract_directory is None:
        relative_extract_directory = pathlib.Path(os.path.splitext(filename)[0])
        extract_directory = get_data_dir().joinpath(relative_extract_directory)
    else:
        extract_directory = pathlib.Path(extract_directory)

    with zipfile.ZipFile(zip_filepath, "r") as zip_file:
        namelist = zip_file.namelist()

        for zipped_filename in namelist:
            if not extract_directory.joinpath(zipped_filename).exists():
                zip_file.extract(zipped_filename, path=extract_directory)

    resolved_paths = [
        extract_directory.joinpath(zipped_filename).resolve()
        for zipped_filename in namelist
    ]

    resolved_filepaths = [path for path in resolved_paths if path.is_file()]

    return resolved_filepaths



# Copyright (C) 2020 Cancer Care Associates

# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at

#     http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


from pymedphys._imports import pymssql
from pymedphys._imports import streamlit as st

import pymedphys


@st.cache(allow_output_mutation=True)
def delivery_from_trf(pandas_table):
    return pymedphys.Delivery._from_pandas(  # pylint: disable = protected-access
        pandas_table
    )


@st.cache(allow_output_mutation=True)
def delivery_from_icom(icom_stream):
    return pymedphys.Delivery.from_icom(icom_stream)


@st.cache(allow_output_mutation=True)
def delivery_from_tel(tel_path):
    return pymedphys.Delivery.from_monaco(tel_path)


@st.cache(hash_funcs={pymssql.Cursor: id}, allow_output_mutation=True)
def delivery_from_mosaiq(cursor_and_field_id):
    cursor, field_id = cursor_and_field_id
    return pymedphys.Delivery.from_mosaiq(cursor, field_id)


def cached_deliveries_loading(inputs, method_function):
    deliveries = []

    for an_input in inputs:
        deliveries += [method_function(an_input)]

    return deliveries



# Copyright (C) 2019 Simon Biggs, Cancer Care Associates

# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at

#     http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import pytest

import matplotlib.pyplot as plt
import numpy as np

from pymedphys._experimental.film import (
    align_images,
    create_image_interpolation,
    interpolated_rotation,
    shift_and_rotate,
)
from pymedphys._mocks.profiles import create_rectangular_field_function


def test_shift_alignment():
    alignment_assertions((5, 5, 0))


def test_rotation_alignment():
    alignment_assertions((0, 0, 20))


@pytest.mark.skip(
    reason="This function is flakey. Not for use until it is able to be reliable"
)
def test_rotation_and_shift_alignment():
    alignment_assertions((-6, 4, -15))


def alignment_assertions(expected):
    ref_field = create_rectangular_field_function((0, 0), (20, 25), 5, rotation=0)

    moving_field = create_rectangular_field_function(
        (-expected[0], -expected[1]), (20, 25), 5, rotation=-expected[2]
    )

    x_span = np.arange(-50, 51)
    y_span = np.arange(-60, 61)
    axes = (x_span, y_span)

    ref_image = ref_field(x_span[:, None], y_span[None, :])
    moving_image = moving_field(x_span[:, None], y_span[None, :])

    x_shift, y_shift, angle = align_images(
        axes, ref_image, axes, moving_image, max_shift=10, max_rotation=30
    )
    shifted_image = shift_and_rotate(axes, axes, moving_image, x_shift, y_shift, angle)

    try:
        # assert np.allclose(expected, (x_shift, y_shift, angle), rtol=0.1, atol=0.1)
        assert np.allclose(shifted_image, ref_image, rtol=0.01, atol=0.01)
    except AssertionError:
        print(x_shift, y_shift, angle)

        plt.figure()
        plt.imshow(ref_image)
        plt.colorbar()

        plt.figure()
        plt.imshow(shifted_image)
        plt.colorbar()

        plt.figure()
        plt.imshow(shifted_image - ref_image)
        plt.colorbar()

        plt.show()
        raise


def test_interpolated_rotation():
    ref_field = create_rectangular_field_function((0, 0), (5, 7), 2, rotation=30)

    moving_field = create_rectangular_field_function((0, 0), (5, 7), 2, rotation=10)

    x_span = np.linspace(-20, 20, 100)
    y_span = np.linspace(-30, 30, 120)

    ref_image = ref_field(x_span[:, None], y_span[None, :])
    moving_image = moving_field(x_span[:, None], y_span[None, :])

    moving_interp = create_image_interpolation((x_span, y_span), moving_image)

    no_rotation = interpolated_rotation(moving_interp, (x_span, y_span), 0)

    try:
        assert np.allclose(no_rotation, moving_image)
    except AssertionError:
        plt.figure()
        plt.imshow(moving_image)
        plt.axis("equal")

        plt.figure()
        plt.imshow(no_rotation)
        plt.axis("equal")

        plt.show()
        raise

    rotated = interpolated_rotation(moving_interp, (x_span, y_span), -20)

    try:
        assert np.allclose(ref_image, rotated, atol=1.0e-1)
    except AssertionError:
        plt.figure()
        plt.imshow(ref_image)
        plt.axis("equal")

        plt.figure()
        plt.imshow(rotated)
        plt.axis("equal")

        plt.figure()
        plt.imshow(rotated - ref_image)
        plt.axis("equal")

        plt.show()
        raise



# -*- coding: utf-8 -*-
from setuptools import setup

package_dir = \
{'': 'lib'}

packages = \
['pymedphys',
 'pymedphys._base',
 'pymedphys._data',
 'pymedphys._dev',
 'pymedphys._dicom',
 'pymedphys._dicom.anonymise',
 'pymedphys._dicom.connect',
 'pymedphys._dicom.constants',
 'pymedphys._dicom.ct',
 'pymedphys._dicom.delivery',
 'pymedphys._dicom.rtplan',
 'pymedphys._dicom.structure',
 'pymedphys._dicom.utilities',
 'pymedphys._electronfactors',
 'pymedphys._experimental',
 'pymedphys._experimental.autosegmentation',
 'pymedphys._experimental.fileformats',
 'pymedphys._experimental.fileformats.mapcheck',
 'pymedphys._experimental.fileformats.mephysto',
 'pymedphys._experimental.fileformats.profiler',
 'pymedphys._experimental.film',
 'pymedphys._experimental.paulking',
 'pymedphys._experimental.pedromartinez',
 'pymedphys._experimental.pedromartinez.oncentra',
 'pymedphys._experimental.pedromartinez.utils',
 'pymedphys._experimental.pinnacle',
 'pymedphys._experimental.plancomplexity',
 'pymedphys._experimental.pseudonymisation',
 'pymedphys._experimental.quickcheck',
 'pymedphys._experimental.serviceplans',
 'pymedphys._gamma',
 'pymedphys._gamma.api',
 'pymedphys._gamma.implementation',
 'pymedphys._gamma.utilities',
 'pymedphys._icom',
 'pymedphys._imports',
 'pymedphys._losslessjpeg',
 'pymedphys._metersetmap',
 'pymedphys._metersetmap.delivery',
 'pymedphys._metersetmap.plt',
 'pymedphys._mocks',
 'pymedphys._monaco',
 'pymedphys._mosaiq',
 'pymedphys._streamlit',
 'pymedphys._streamlit.apps',
 'pymedphys._streamlit.apps.metersetmap',
 'pymedphys._streamlit.utilities',
 'pymedphys._trf',
 'pymedphys._trf.decode',
 'pymedphys._trf.manage',
 'pymedphys._utilities',
 'pymedphys._utilities.algorithms',
 'pymedphys._utilities.constants',
 'pymedphys._utilities.filehash',
 'pymedphys._utilities.transforms',
 'pymedphys._vendor',
 'pymedphys._vendor.apipkg',
 'pymedphys._vendor.pylinac',
 'pymedphys._vendor.pylinac.core',
 'pymedphys._wlutz',
 'pymedphys.beta',
 'pymedphys.cli',
 'pymedphys.cli.experimental',
 'pymedphys.docs',
 'pymedphys.experimental',
 'pymedphys.tests',
 'pymedphys.tests.coordinates',
 'pymedphys.tests.delivery',
 'pymedphys.tests.dicom',
 'pymedphys.tests.dicom.rtplan',
 'pymedphys.tests.e2e',
 'pymedphys.tests.experimental',
 'pymedphys.tests.experimental.film',
 'pymedphys.tests.experimental.mapcheck',
 'pymedphys.tests.experimental.mephysto',
 'pymedphys.tests.experimental.paulking',
 'pymedphys.tests.experimental.paulking.film',
 'pymedphys.tests.experimental.paulking.test_coll',
 'pymedphys.tests.experimental.pinnacle',
 'pymedphys.tests.experimental.profiler',
 'pymedphys.tests.experimental.pseudonymisation',
 'pymedphys.tests.gamma',
 'pymedphys.tests.logfiles',
 'pymedphys.tests.logging',
 'pymedphys.tests.losslessjpeg',
 'pymedphys.tests.metersetmap',
 'pymedphys.tests.mocks',
 'pymedphys.tests.monaco',
 'pymedphys.tests.trf',
 'pymedphys.tests.utilities',
 'pymedphys.tests.winstonlutz']

package_data = \
{'': ['*'],
 'pymedphys._experimental': ['serviceplans/templates/*'],
 'pymedphys._streamlit': ['apps/data/*'],
 'pymedphys.docs': ['_static/*',
                    '_templates/*',
                    'background/*',
                    'howto/*',
                    'howto/advanced/*',
                    'howto/gamma/*',
                    'img/*',
                    'ref/*',
                    'ref/cli/*',
                    'ref/gui/*',
                    'ref/lib/*',
                    'ref/lib/experimental/*',
                    'tutes/*'],
 'pymedphys.tests': ['dicom/data/rtplan/*',
                     'dicom/scratch/*',
                     'e2e/cypress/*',
                     'e2e/cypress/fixtures/*',
                     'e2e/cypress/integration/streamlit/*',
                     'e2e/cypress/plugins/*',
                     'e2e/cypress/support/*',
                     'experimental/mephysto/data/baselines/*',
                     'experimental/mephysto/data/measurements/*',
                     'experimental/paulking/film/data/*']}

extras_require = \
{':python_version >= "3.6.0" and python_version < "3.7.0"': ['dataclasses'],
 'dev': ['tqdm',
         'attrs',
         'watchdog',
         'keyring',
         'packaging',
         'PyYAML',
         'requests',
         'numpy>=1.12',
         'matplotlib',
         'scipy',
         'pandas',
         'Pillow',
         'imageio',
         'shapely>=1.7.0',
         'dbfread',
         'pydicom>=2.0.0',
         'pynetdicom',
         'pylibjpeg-libjpeg',
         'immutables',
         'cython',
         'pymssql',
         'kiwisolver!=1.3.0',
         'python_dateutil',
         'scikit-image',
         'streamlit==0.70.0',
         'timeago',
         'sphinx>=1.4,<1.8',
         'sphinx-rtd-theme>=0.4.3,<0.5.0',
         'sphinxcontrib-napoleon',
         'sphinx-argparse',
         'sphinx-autobuild',
         'nbsphinx',
         'jupyter_client',
         'ipython',
         'ipykernel',
         'nbconvert==6.0.6',
         'pytest',
         'pytest-sugar',
         'hypothesis',
         'psutil',
         'pylint',
         'pytest-pylint',
         'pre-commit',
         'black>=19.3b0,<20.0',
         'mypy',
         'rope',
         'doc8',
         'tomlkit'],
 'dicom': ['pydicom>=2.0.0', 'pynetdicom', 'pylibjpeg-libjpeg', 'immutables'],
 'docs': ['sphinx>=1.4,<1.8',
          'sphinx-rtd-theme>=0.4.3,<0.5.0',
          'sphinxcontrib-napoleon',
          'sphinx-argparse',
          'sphinx-autobuild',
          'nbsphinx',
          'jupyter_client',
          'ipython',
          'ipykernel',
          'nbconvert==6.0.6'],
 'doctests': ['tensorflow>=2.2.0', 'black>=19.3b0,<20.0', 'tomlkit'],
 'experimental': ['python_dateutil', 'scikit-image'],
 'gui': ['streamlit==0.70.0', 'timeago'],
 'ml': ['tensorflow>=2.2.0'],
 'tests': ['pytest',
           'pytest-sugar',
           'hypothesis',
           'psutil',
           'pylint',
           'pytest-pylint'],
 'user': ['tqdm',
          'attrs',
          'watchdog',
          'keyring',
          'packaging',
          'PyYAML',
          'requests',
          'numpy>=1.12',
          'matplotlib',
          'scipy',
          'pandas',
          'Pillow',
          'imageio',
          'shapely>=1.7.0',
          'dbfread',
          'pydicom>=2.0.0',
          'pynetdicom',
          'pylibjpeg-libjpeg',
          'immutables',
          'cython',
          'pymssql',
          'kiwisolver!=1.3.0',
          'python_dateutil',
          'scikit-image',
          'streamlit==0.70.0',
          'timeago']}

entry_points = \
{'console_scripts': ['pymedphys = pymedphys.cli:pymedphys_cli']}

setup_kwargs = {
    'name': 'pymedphys',
    'version': '0.34.0.dev12',
    'description': 'Medical Physics library',
    'long_description': "|logo|\n\n.. |logo| image:: https://github.com/pymedphys/pymedphys/raw/3f8d82fc3b53eb636a75336477734e39fa406110/docs/logos/pymedphys_title.png\n    :target: https://docs.pymedphys.com/\n\n.. START_OF_DOCS_IMPORT\n\n**A community effort to develop an open standard library for Medical Physics\nin Python. Building quality transparent software together via peer review\nand open source distribution. Open code is better science.**\n\n|build| |pypi| |python| |license|\n\n.. |build| image:: https://img.shields.io/endpoint.svg?url=https%3A%2F%2Factions-badge.atrox.dev%2Fpymedphys%2Fpymedphys%2Fbadge&label=build&logo=none\n    :target: https://actions-badge.atrox.dev/pymedphys/pymedphys/goto\n\n.. |pypi| image:: https://img.shields.io/pypi/v/pymedphys\n    :target: https://pypi.org/project/pymedphys/\n\n.. |python| image:: https://img.shields.io/pypi/pyversions/pymedphys\n    :target: https://pypi.org/project/pymedphys/\n\n.. |license| image:: https://img.shields.io/pypi/l/pymedphys\n    :target: https://choosealicense.com/licenses/apache-2.0/\n\n\nWhat is PyMedPhys?\n------------------\n\nAn open-source Medical Physics python library with a focus on being\na place to share, review, improve, and transparently learn off of each\nother's code. It is inspired by the collaborative work of our physics peers\nin astronomy and their `Astropy Project`_. PyMedPhys is available on `PyPI`_\nand `GitHub`_.\n\n.. _`Astropy Project`: http://www.astropy.org/\n.. _`PyPI`: https://pypi.org/project/pymedphys/\n.. _`GitHub`: https://github.com/pymedphys/pymedphys\n\n\nMailing list\n------------\n\nIf you would like to dive into the community a great place to get started is\nto sign up to `the mailing list`_ and say hi by introducing yourself with\nwhere you're from, and what you hope to achieve with PyMedPhys.\n\n.. _`the mailing list`: https://groups.google.com/g/pymedphys\n\n\nTable of contents\n-----------------\n\n`Tutorials`_\n........................\n\nGet started with a hands-on introduction to PyMedPhys for beginners\n\n`How-To guides`_\n........................\n\nGuides and recipes for common problems and tasks\n\n`Reference`_\n............................\n\nTechnical reference for the `library`_ (modules, functions and classes),\nas well as the available `command line tools`_.\n\n`Background`_\n..............................\n\nExplanation and discussion of key topics and concepts\n\n\nBeta level of development\n-------------------------\n\nPyMedPhys is currently within the ``beta`` stage of its life-cycle. It will\nstay in this stage until the version number leaves ``0.x.x`` and enters\n``1.x.x``. While PyMedPhys is in ``beta`` stage, **no API is guaranteed to be\nstable from one release to the next.** In fact, it is very likely that the\nentire API will change multiple times before a ``1.0.0`` release. In practice,\nthis means that upgrading ``pymedphys`` to a new version will possibly break\nany code that was using the old version of pymedphys. We try to be abreast of\nthis by providing details of any breaking changes from one release to the next\nwithin the `Release Notes`_.\n\nOur Team\n--------\n\nPyMedPhys is what it is today due to its maintainers and contributors both past\nand present. Here is our team.\n\nMaintainers\n...........\n\n* `Simon Biggs`_\n    * `Riverina Cancer Care Centre`_, Australia\n\n.. _`Simon Biggs`: https://github.com/SimonBiggs\n\n* `Stuart Swerdloff`_\n    * `ELEKTA Pty Ltd`_: New Zealand\n\n.. _`Stuart Swerdloff`: https://github.com/sjswerdloff\n\n|rccc| |sjs|\n\nActive contributors\n...................\n\n* `Phillip Chlap`_\n    * `University of New South Wales`_, Australia\n    * `Ingham Institute`_, Australia\n\n.. _`Phillip Chlap`: https://github.com/pchlap\n\n* `Matthew Cooper`_\n\n.. _`Matthew Cooper`: https://github.com/matthewdeancooper\n\n* `Jake Rembish`_\n    * `UT Health San Antonio`_, USA\n\n.. _`Jake Rembish`: https://github.com/rembishj\n\n* `Pedro Martinez`_\n    * `University of Calgary`_, Canada\n    * `Tom Baker Cancer Centre`_, Canada\n\n.. _`Pedro Martinez`: https://github.com/peterg1t\n\n* `Rafael Ayala`_\n    * `Hospital General Universitario Gregorio Marañón`_, Spain\n\n.. _`Rafael Ayala`: https://github.com/ayalalazaro\n\n\n|uth| |uoc| |hgugm|\n\n\nMaintainer emeritus\n...................\n\n* `Matthew Jennings`_\n    * `Royal Adelaide Hospital`_, Australia\n\n.. _`Matthew Jennings`: https://github.com/Matthew-Jennings\n\n|rah|\n\nPast contributors\n.................\n\n* `Matthew Sobolewski <https://github.com/msobolewski>`_\n* `Paul King <https://github.com/kingrpaul>`_\n* `Jacob McAloney <https://github.com/JacobMcAloney>`_\n\n\n.. |rccc| image:: https://github.com/pymedphys/pymedphys/raw/3f8d82fc3b53eb636a75336477734e39fa406110/docs/logos/rccc_200x200.png\n    :target: `Riverina Cancer Care Centre`_\n\n.. |rah| image:: https://github.com/pymedphys/pymedphys/raw/3f8d82fc3b53eb636a75336477734e39fa406110/docs/logos/gosa_200x200.png\n    :target: `Royal Adelaide Hospital`_\n\n.. |jarmc| image:: https://github.com/pymedphys/pymedphys/raw/3f8d82fc3b53eb636a75336477734e39fa406110/docs/logos/jarmc_200x200.png\n    :target: `Anderson Regional Cancer Center`_\n\n.. |nbcc| image:: https://github.com/pymedphys/pymedphys/raw/3f8d82fc3b53eb636a75336477734e39fa406110/docs/logos/nbcc_200x200.png\n    :target: `Northern Beaches Cancer Care`_\n\n.. |uoc| image:: https://github.com/pymedphys/pymedphys/raw/3f8d82fc3b53eb636a75336477734e39fa406110/docs/logos/uoc_200x200.png\n    :target: `University of Calgary`_\n\n.. |uth| image:: https://github.com/pymedphys/pymedphys/raw/3f8d82fc3b53eb636a75336477734e39fa406110/docs/logos/UTHSA_logo.png\n    :target: `UT Health San Antonio`_\n\n.. |hgugm| image:: https://github.com/pymedphys/pymedphys/raw/3f8d82fc3b53eb636a75336477734e39fa406110/docs/logos/HGUGM_200x200.png\n    :target: `Hospital General Universitario Gregorio Marañón`_\n\n.. |sjs| image:: https://github.com/pymedphys/pymedphys/raw/7e9204656e0468b0843533472553a03a99387386/logos/swerdloff.png\n    :target: `Swerdloff Family`_\n\n.. _`Riverina Cancer Care Centre`: https://www.riverinacancercare.com.au/\n\n.. _`ELEKTA Pty Ltd`: https://www.elekta.com/\n\n.. _`Royal Adelaide Hospital`: https://www.rah.sa.gov.au/\n\n.. _`University of New South Wales`: https://www.unsw.edu.au/\n\n.. _`South Western Sydney Local Health District`: https://www.swslhd.health.nsw.gov.au/\n\n.. _`Anderson Regional Cancer Center`: https://www.andersonregional.org/services/cancer-care/\n\n.. _`Northern Beaches Cancer Care`: https://www.northernbeachescancercare.com.au/\n\n.. _`University of Calgary`: https://www.ucalgary.ca/\n\n.. _`Tom Baker Cancer Centre`: https://www.ahs.ca/tbcc\n\n.. _`UT Health San Antonio`: https://www.uthscsa.edu/academics/biomedical-sciences/programs/radiological-sciences-phd\n\n.. _`Hospital General Universitario Gregorio Marañón`: https://www.comunidad.madrid/hospital/gregoriomaranon/\n\n.. _`Swerdloff Family`: https://github.com/sjswerdloff\n\n.. _`Ingham Institute`: https://inghaminstitute.org.au/\n\n.. END_OF_DOCS_IMPORT\n\n.. _`Tutorials`: https://docs.pymedphys.com/tutes\n.. _`How-To guides`: https://docs.pymedphys.com/howto\n.. _`Reference`: https://docs.pymedphys.com/ref\n.. _`Background`: https://docs.pymedphys.com/background\n\n.. _`library`: https://docs.pymedphys.com/ref/lib\n.. _`command line tools`: https://docs.pymedphys.com/ref/cli\n\n.. _`Release Notes`: http://docs.pymedphys.com/release-notes.html\n",
    'author': 'PyMedPhys Contributors',
    'author_email': 'developers@pymedphys.com',
    'maintainer': None,
    'maintainer_email': None,
    'url': 'https://pymedphys.com',
    'package_dir': package_dir,
    'packages': packages,
    'package_data': package_data,
    'extras_require': extras_require,
    'entry_points': entry_points,
    'python_requires': '>=3.6,<4.0',
}


setup(**setup_kwargs)



# Copyright (C) 2019 Simon Biggs

# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at

#     http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


from pymedphys._base.delivery import DeliveryBase

from pymedphys._vendor.deprecated import deprecated as _deprecated

from ..metersetmap import calc_metersetmap


class DeliveryMetersetMap(DeliveryBase):
    def metersetmap(
        self,
        gantry_angles=None,
        gantry_tolerance=3,
        grid_resolution=None,
        max_leaf_gap=None,
        leaf_pair_widths=None,
        min_step_per_pixel=None,
        output_always_list=False,
    ):
        if gantry_angles is None:
            gantry_angles = 0
            gantry_tolerance = 500
        else:
            gantry_angles = tuple(gantry_angles)

        filtered_delivery = self._filter_cps()

        masked_by_gantry = filtered_delivery._mask_by_gantry(  # pylint: disable = protected-access
            gantry_angles, gantry_tolerance
        )

        metersetmaps = []
        for delivery_data in masked_by_gantry:
            metersetmaps.append(
                calc_metersetmap(
                    delivery_data.monitor_units,
                    delivery_data.mlc,
                    delivery_data.jaw,
                    grid_resolution=grid_resolution,
                    max_leaf_gap=max_leaf_gap,
                    leaf_pair_widths=leaf_pair_widths,
                    min_step_per_pixel=min_step_per_pixel,
                )
            )

        if not output_always_list:
            if len(metersetmaps) == 1:
                return metersetmaps[0]

        return metersetmaps

    @_deprecated(
        reason=(
            "pymedphys.Delivery.mudensity has been replaced by "
            "pymedphys.Delivery.metersetmap"
        )
    )
    def mudensity(
        self,
        gantry_angles=None,
        gantry_tolerance=3,
        grid_resolution=None,
        max_leaf_gap=None,
        leaf_pair_widths=None,
        min_step_per_pixel=None,
        output_always_list=False,
    ):
        return self.metersetmap(
            gantry_angles=gantry_angles,
            gantry_tolerance=gantry_tolerance,
            grid_resolution=grid_resolution,
            max_leaf_gap=max_leaf_gap,
            leaf_pair_widths=leaf_pair_widths,
            min_step_per_pixel=min_step_per_pixel,
            output_always_list=output_always_list,
        )



# The MIT License (MIT)

# Copyright (c) 2017 Laurent LAPORTE

# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:

# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.

# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

# Vendored with modifications from https://stackoverflow.com/a/40301488

import functools
import inspect
import warnings

string_types = (type(b""), type(u""))


def deprecated(reason):
    """
    This is a decorator which can be used to mark functions
    as deprecated. It will result in a warning being emitted
    when the function is used.
    """

    if isinstance(reason, string_types):

        # The @deprecated is used with a 'reason'.
        #
        # .. code-block:: python
        #
        #    @deprecated("please, use another function")
        #    def old_function(x, y):
        #        pass

        def decorator(func1):

            if inspect.isclass(func1):
                fmt1 = "Call to deprecated class `{name}` ({reason})."
            else:
                fmt1 = "Call to deprecated function `{name}` ({reason})."

            @functools.wraps(func1)
            def new_func1(*args, **kwargs):
                warnings.warn(
                    fmt1.format(name=func1.__name__, reason=reason),
                    category=DeprecationWarning,
                    stacklevel=2,
                )
                return func1(*args, **kwargs)

            return new_func1

        return decorator

    elif inspect.isclass(reason) or inspect.isfunction(reason):

        # The @deprecated is used without any 'reason'.
        #
        # .. code-block:: python
        #
        #    @deprecated
        #    def old_function(x, y):
        #        pass

        func2 = reason

        if inspect.isclass(func2):
            fmt2 = "Call to deprecated class `{name}`."
        else:
            fmt2 = "Call to deprecated function `{name}`."

        @functools.wraps(func2)
        def new_func2(*args, **kwargs):
            warnings.warn(
                fmt2.format(name=func2.__name__),
                category=DeprecationWarning,
                stacklevel=2,
            )
            return func2(*args, **kwargs)

        return new_func2

    else:
        raise TypeError(repr(type(reason)))



# Copyright (C) 2018 Cancer Care Associates

# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at

#     http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from typing import Type

from pymedphys._imports import numpy as np

from pymedphys._base.delivery import DeliveryBase, DeliveryGeneric

from .constants import (
    COLLIMATOR_NAME,
    GANTRY_NAME,
    JAW_NAMES,
    Y1_LEAF_BANK_NAMES,
    Y2_LEAF_BANK_NAMES,
)
from .trf2pandas import read_trf


class DeliveryLogfile(DeliveryBase):
    @classmethod
    def from_logfile(cls, filepath):
        _, dataframe = read_trf(filepath)

        return cls._from_pandas(dataframe)

    @classmethod
    def _from_pandas(cls: Type[DeliveryGeneric], table) -> DeliveryGeneric:
        raw_monitor_units = table["Step Dose/Actual Value (Mu)"]

        diff = np.append([0], np.diff(raw_monitor_units))
        diff[diff < 0] = 0

        monitor_units = np.cumsum(diff)

        gantry = table[GANTRY_NAME]
        collimator = table[COLLIMATOR_NAME]

        y1_bank = [table[name] for name in Y1_LEAF_BANK_NAMES]

        y2_bank = [table[name] for name in Y2_LEAF_BANK_NAMES]

        mlc = [y1_bank, y2_bank]
        mlc = np.swapaxes(mlc, 0, 2)

        jaw = [table[name] for name in JAW_NAMES]
        jaw = np.swapaxes(jaw, 0, 1)

        return cls(monitor_units, gantry, collimator, mlc, jaw)



# Copyright (C) 2019 South Western Sydney Local Health District,
# University of New South Wales

# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at

#     http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

# This work is derived from:
# https://github.com/AndrewWAlexander/Pinnacle-tar-DICOM
# which is released under the following license:

# Copyright (c) [2017] [Colleen Henschel, Andrew Alexander]

# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:

# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.

# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

# pylint: disable = redefined-outer-name

import os
import tempfile
from zipfile import ZipFile

import pytest

import numpy as np

import pydicom

from pymedphys._data import download
from pymedphys.experimental.pinnacle import PinnacleExport

working_path = tempfile.mkdtemp()
data_path = os.path.join(working_path, "data")


def get_online_data(filename):
    return download.get_file_within_data_zip("pinnacle_test_data.zip", filename)


@pytest.fixture(scope="session")
def data():

    zip_ref = ZipFile(get_online_data("pinnacle_16.0_test_data.zip"), "r")
    zip_ref.extractall(data_path)
    zip_ref.close()

    return data_path


@pytest.fixture
def pinn(data):

    pinn_objs = []

    for d in os.listdir(data):
        pinn_dir = os.path.join(data, d, "Pinnacle")
        for pat_dir in os.listdir(pinn_dir):
            pinn_objs.append(PinnacleExport(os.path.join(pinn_dir, pat_dir), None))

    return pinn_objs


@pytest.mark.slow
def test_pinnacle(pinn):

    for p in pinn:
        plans = p.plans
        assert len(plans) == 1


def find_corresponding_dicom(dcm):

    for root, _, files in os.walk(data_path):

        dicom_files = [f for f in files if f.endswith(".dcm")]
        for f in dicom_files:

            dcm_file = os.path.join(root, f)
            ds = pydicom.read_file(dcm_file)

            if ds.PatientID == dcm.PatientID and ds.Modality == dcm.Modality:

                if ds.Modality == "CT":
                    # Also match the SliceLocation
                    if not ds.SliceLocation == dcm.SliceLocation:
                        continue
                return ds

    return None


@pytest.mark.slow
@pytest.mark.pydicom
def test_ct(pinn):

    for p in pinn:
        export_path = os.path.join(
            working_path, "output", p.patient_info["MedicalRecordNumber"], "CT"
        )
        os.makedirs(export_path)

        export_plan = p.plans[0]

        p.export_image(export_plan.primary_image, export_path=export_path)

        # Get the exported CT file
        for f in os.listdir(export_path):
            if f.startswith("CT"):
                exported_ct = pydicom.read_file(os.path.join(export_path, f))
                assert exported_ct.Modality == "CT"

                # Get the ground truth CT file
                pinn_ct = find_corresponding_dicom(exported_ct)
                assert pinn_ct is not None

                # Some (very) basic sanity checks
                assert pinn_ct.PatientID == exported_ct.PatientID
                assert pinn_ct.SliceLocation == exported_ct.SliceLocation

                # Get the image data
                exported_img = exported_ct.pixel_array.astype(np.int16)
                pinn_img = pinn_ct.pixel_array.astype(np.int16)

                # Ensure images are the same size
                assert exported_img.shape == pinn_img.shape

                # Make sure the absolute difference is (close to) zero
                assert np.allclose(exported_img, pinn_img, atol=0.00001)


@pytest.mark.slow
@pytest.mark.pydicom
def test_struct(pinn):

    for p in pinn:
        export_path = os.path.join(
            working_path, "output", p.patient_info["MedicalRecordNumber"], "RTSTRUCT"
        )
        os.makedirs(export_path)

        export_plan = p.plans[0]

        p.export_struct(export_plan, export_path=export_path)

        # Get the exported struct file
        for f in os.listdir(export_path):
            if f.startswith("RS"):
                exported_struct = pydicom.read_file(os.path.join(export_path, f))
                assert exported_struct.Modality == "RTSTRUCT"

        # Get the ground truth RTSTRUCT file
        pinn_struct = find_corresponding_dicom(exported_struct)
        assert pinn_struct is not None

        assert len(pinn_struct.StructureSetROISequence) == len(
            exported_struct.StructureSetROISequence
        )

        for s1, s2 in zip(
            pinn_struct.StructureSetROISequence, exported_struct.StructureSetROISequence
        ):
            assert s1.ROIName == s2.ROIName

        # TODO: Generate a mask for each contour and do a comparison of these


def assert_same_dose(dose_a, dose_b):

    # Check the same patient orientation
    assert dose_a.ImageOrientationPatient == dose_b.ImageOrientationPatient

    # Get the dose volumes
    exported_vol = dose_a.pixel_array.astype(np.int16)
    pinn_vol = dose_b.pixel_array.astype(np.int16)

    # Ensure dose volumes are the same size
    assert exported_vol.shape == pinn_vol.shape

    # Apply dose grid scaling
    exported_vol = exported_vol * dose_a.DoseGridScaling
    pinn_vol = pinn_vol * dose_b.DoseGridScaling

    # Make sure the maximum values are in the same locations
    assert exported_vol.argmax() == pinn_vol.argmax()

    # Make sure the absolute difference is (close to) zero
    assert np.allclose(exported_vol, pinn_vol, atol=0.01)


@pytest.mark.slow
@pytest.mark.pydicom
def test_dose(pinn):

    for p in pinn:
        export_path = os.path.join(
            working_path, "output", p.patient_info["MedicalRecordNumber"], "RTDOSE"
        )
        os.makedirs(export_path)

        export_plan = p.plans[0]

        p.export_dose(export_plan, export_path)

        # Get the exported RTDOSE file
        for f in os.listdir(export_path):
            if f.startswith("RD"):
                exported_dose = pydicom.read_file(os.path.join(export_path, f))
                assert exported_dose.Modality == "RTDOSE"
                break

        # Get the ground truth RTDOSE file
        pinn_dose = find_corresponding_dicom(exported_dose)
        assert pinn_dose is not None

        assert_same_dose(pinn_dose, exported_dose)


@pytest.mark.slow
@pytest.mark.pydicom
def test_plan(pinn):

    for p in pinn:
        export_path = os.path.join(
            working_path, "output", p.patient_info["MedicalRecordNumber"], "RTPLAN"
        )
        os.makedirs(export_path)

        export_plan = p.plans[0]

        p.export_plan(export_plan, export_path)

        # Get the exported RTPLAN file
        for f in os.listdir(export_path):
            if f.startswith("RP"):
                exported_plan = pydicom.read_file(os.path.join(export_path, f))
                assert exported_plan.Modality == "RTPLAN"
                break

        # Get the ground truth RTDOSE file
        pinn_plan = find_corresponding_dicom(exported_plan)
        assert pinn_plan is not None

        assert pinn_plan.RTPlanName == exported_plan.RTPlanName
        assert pinn_plan.RTPlanLabel == exported_plan.RTPlanLabel

        # TODO The RTPLAN export isn't fully functional yet, so we need
        # to test more as we add that functionality


@pytest.mark.slow
def test_missing_image():

    zip_ref = ZipFile(get_online_data("pinnacle_test_data_no_image.zip"), "r")
    zip_ref.extractall(data_path)
    zip_ref.close()

    pinn_path = os.path.join(data_path, "Pt3", "Pinnacle", "Patient_7430")
    pinn = PinnacleExport(pinn_path, None)

    # There is no image to export, so this line should fail gracefully and simply output
    # a warning to the logger
    pinn.export_image(pinn.plans[0].primary_image, export_path=".")


@pytest.fixture(scope="session")
def orientation_data():

    zip_ref = ZipFile(get_online_data("pinnacle_orientations_test_data.zip"), "r")

    zip_ref.extractall(data_path)
    zip_ref.close()

    return data_path


@pytest.fixture
def orientation_pinn(orientation_data):

    pinn_path = os.path.join(orientation_data, "pinnacle", "Patient_22902")

    pinn_obj = PinnacleExport(pinn_path, None)

    return pinn_obj


@pytest.mark.slow
@pytest.mark.pydicom
def test_dose_hfs(orientation_pinn):
    # Test exporting dose for patients in HFS position

    export_path = os.path.join(
        working_path, "output", "pinn_orientations", "RTDOSE", "HFS"
    )
    os.makedirs(export_path)

    export_plan = orientation_pinn.plans[0]

    orientation_pinn.export_dose(export_plan, export_path)

    # Get the exported RTDOSE file
    for f in os.listdir(export_path):
        if f.startswith("RD"):
            exported_dose = pydicom.read_file(os.path.join(export_path, f))
            assert exported_dose.Modality == "RTDOSE"
            break

    # Get the ground truth RTDOSE file
    gt_dose_file = "1.3.46.670589.13.997910418.20200707132626.548267.dcm"
    gt_dose_path = os.path.join(data_path, "dicom", "HFS", gt_dose_file)
    pinn_dose = pydicom.read_file(gt_dose_path)
    assert pinn_dose is not None

    assert_same_dose(exported_dose, pinn_dose)


@pytest.mark.slow
@pytest.mark.pydicom
def test_dose_hfp(orientation_pinn):
    # Test exporting dose for patients in HFP position

    export_path = os.path.join(
        working_path, "output", "pinn_orientations", "RTDOSE", "HFP"
    )
    os.makedirs(export_path)

    export_plan = orientation_pinn.plans[1]

    orientation_pinn.export_dose(export_plan, export_path)

    # Get the exported RTDOSE file
    for f in os.listdir(export_path):
        if f.startswith("RD"):
            exported_dose = pydicom.read_file(os.path.join(export_path, f))
            assert exported_dose.Modality == "RTDOSE"
            break

    # Get the ground truth RTDOSE file
    gt_dose_file = "1.3.46.670589.13.997910418.20200707132247.647011.dcm"
    gt_dose_path = os.path.join(data_path, "dicom", "HFP", gt_dose_file)
    pinn_dose = pydicom.read_file(gt_dose_path)
    assert pinn_dose is not None

    assert_same_dose(exported_dose, pinn_dose)


@pytest.mark.slow
@pytest.mark.pydicom
def test_dose_ffs(orientation_pinn):
    # Test exporting dose for patients in FFS position

    export_path = os.path.join(
        working_path, "output", "pinn_orientations", "RTDOSE", "FFS"
    )
    os.makedirs(export_path)

    export_plan = orientation_pinn.plans[2]

    orientation_pinn.export_dose(export_plan, export_path)

    # Get the exported RTDOSE file
    for f in os.listdir(export_path):
        if f.startswith("RD"):
            exported_dose = pydicom.read_file(os.path.join(export_path, f))
            assert exported_dose.Modality == "RTDOSE"
            break

    # Get the ground truth RTDOSE file
    gt_dose_file = "1.3.46.670589.13.997910418.20200707132449.572508.dcm"
    gt_dose_path = os.path.join(data_path, "dicom", "FFS", gt_dose_file)
    pinn_dose = pydicom.read_file(gt_dose_path)
    assert pinn_dose is not None

    assert_same_dose(exported_dose, pinn_dose)


@pytest.mark.slow
@pytest.mark.pydicom
def test_dose_ffp(orientation_pinn):
    # Test exporting dose for patients in FFP position

    export_path = os.path.join(
        working_path, "output", "pinn_orientations", "RTDOSE", "FFP"
    )
    os.makedirs(export_path)

    export_plan = orientation_pinn.plans[3]

    orientation_pinn.export_dose(export_plan, export_path)

    # Get the exported RTDOSE file
    for f in os.listdir(export_path):
        if f.startswith("RD"):
            exported_dose = pydicom.read_file(os.path.join(export_path, f))
            assert exported_dose.Modality == "RTDOSE"
            break

    # Get the ground truth RTDOSE file
    gt_dose_file = "1.3.46.670589.13.997910418.20200707132737.923688.dcm"
    gt_dose_path = os.path.join(data_path, "dicom", "FFP", gt_dose_file)
    pinn_dose = pydicom.read_file(gt_dose_path)
    assert pinn_dose is not None

    assert_same_dose(exported_dose, pinn_dose)



# Copyright (C) 2020 Simon Biggs

# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at

#     http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


from pymedphys._imports import numpy as np
from pymedphys._imports import skimage


def soft_surface_dice(reference, evaluation):
    """Non-TensorFlow implementation of a soft surface dice
    """
    edge_reference = skimage.filters.scharr(reference)
    edge_evaluation = skimage.filters.scharr(evaluation)

    score = np.sum(np.abs(edge_evaluation - edge_reference)) / np.sum(
        edge_evaluation + edge_reference
    )

    return 1 - score


def tf_soft_surface_dice(reference, evaluation):
    """TensorFlow implementation of a soft surface dice
    """
    raise NotImplementedError()



ICOM = {
    "Patient ID": (b" \x00LO\x00P", str, "first"),
    "Patient Name": (b"\x10\x00PN\x00P", str, "first"),
    "Rx Site": (b"\x01\x10LO\x00P", str, "first"),
    "Field ID": (b"\x03\x10LO\x00P", str, "first"),
    "Machine ID": (b"\xb2\x00SH\x00P", str, "first"),
    "Radiation Type": (b"\xc6\x00CS\x00R", str, "first"),
    "Energy": (b"\x14\x01SH\x00R", str, "first"),
    "Wedge": (b"\x18\x01CS\x00R", str, "first"),
    "Segment": (b"\x07\x10DS\x00R", int, "first"),
    "Total MU": (b"\t\x10DS\x00R", float, "first"),
    "Delivery MU": (b"2\x00DS\x00R", float, "first"),
    "Backup Delivery MU": (b"3\x00DS\x00R", float, "first"),
    "Beam Timer": (b"8\x00SH\x00R", float, "first"),
    "Segment MU": (b"\x0b\x00DS\x00R", float, "first"),
    "Gantry": (b"\x1e\x01DS\x00R", float, "first"),
    "Collimator": (b" \x01DS\x00R", float, "first"),
    "Table Column": (b'"\x01DS\x00R', int, "first"),
    "Table Isocentric": (rb"\%\x01DS\x00R", int, "first"),
    "Table Vertical": (rb"\(\x01DS\x00R", float, "first"),
    "Table Longitudinal": (rb"\)\x01DS\x00R", float, "first"),
    "Table Lateral": (rb"\*\x01DS\x00R", float, "first"),
    "Beam Description": (b"\x0c\x00SH\x00R", str, "first"),
    "Interlocks": (b"\x16\x10LO\x00R", str, "all"),
    "Previous Interlocks": (b"\x18\x10LO\x00R", str, "all"),
}



# Copyright (C) 2018 Cancer Care Associates

# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at

#     http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from pymedphys._imports import numpy as np


def find_relevant_control_points(mu):
    """Returns that control points that had an MU difference either side.
    """
    mu_diff = np.diff(mu)
    no_change = mu_diff == 0
    try:
        start = no_change[0]
        end = no_change[-1]
    except IndexError:
        all_true = np.empty_like(mu).astype(bool)  # pylint: disable = no-member
        all_true.fill(True)
        return all_true

    no_change_before = no_change[0:-1]
    no_change_after = no_change[1::]

    no_change_before_and_after = no_change_before & no_change_after

    irrelevant_control_point = np.hstack([start, no_change_before_and_after, end])
    relevant_control_points = np.invert(irrelevant_control_point)

    return relevant_control_points


def remove_irrelevant_control_points(monitor_units, *args):
    relevant_control_points = find_relevant_control_points(monitor_units)

    provided_values = tuple((monitor_units, *args))

    result = tuple(np.array(item)[relevant_control_points] for item in provided_values)

    return result


def to_tuple(a):
    # https://stackoverflow.com/a/10016613/3912576
    try:
        return tuple(to_tuple(i) for i in a)
    except TypeError:
        return a



# Copyright (C) 2018 Cancer Care Associates

# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at

#     http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Create a reproducible hash from a filepath.
"""

import hashlib


def hash_file(filename, dot_feedback=False):
    BLOCKSIZE = 65536
    hasher = hashlib.sha1()
    with open(filename, "rb") as afile:
        buf = afile.read(BLOCKSIZE)
        while len(buf) > 0:
            hasher.update(buf)
            buf = afile.read(BLOCKSIZE)

    if dot_feedback:
        print(".", end="", flush=True)

    return hasher.hexdigest()



# Copyright (C) 2018 Simon Biggs

# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at

#     http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


"""The tests given here are replicated using pymedphys.gamma from the
method given within the following paper:

> C. Agnew, C. McGarry, A tool to include gamma analysis software into a
> quality assurance program. Radiotherapy and Oncology (2016), >
http://dx.doi.org/10.1016/j.radonc.2015.11.034
"""

import pytest

import numpy as np

import pydicom

import pymedphys
from pymedphys._data import download
from pymedphys._gamma.utilities import calculate_pass_rate

# pylint: disable=C0103,C1801


def get_data_file(filename):
    return download.get_file_within_data_zip("gamma_test_data.zip", filename)


RANDOM_SUBSET = 50000


def dose_from_dataset(ds):
    pixels = ds.pixel_array
    dose = pixels * ds.DoseGridScaling

    return dose


def load_yx_from_dicom(ds):
    resolution = np.array(ds.PixelSpacing).astype(float)
    dx = resolution[0]
    x = ds.ImagePositionPatient[0] + np.arange(0, ds.Columns * dx, dx)

    dy = resolution[1]
    y = ds.ImagePositionPatient[1] + np.arange(0, ds.Rows * dy, dy)

    return y, x


def run_gamma(
    filepath_ref,
    filepath_eval,
    random_subset=None,
    max_gamma=1.1,
    dose_threshold=1,
    distance_threshold=1,
):

    if random_subset is not None:
        np.random.seed(42)

    ds_ref = pydicom.read_file(filepath_ref)
    ds_eval = pydicom.read_file(filepath_eval)

    axes_reference = load_yx_from_dicom(ds_ref)
    dose_reference = dose_from_dataset(ds_ref)

    axes_evaluation = load_yx_from_dicom(ds_eval)
    dose_evaluation = dose_from_dataset(ds_eval)

    gamma = pymedphys.gamma(
        axes_reference,
        dose_reference,
        axes_evaluation,
        dose_evaluation,
        dose_threshold,
        distance_threshold,
        lower_percent_dose_cutoff=20,
        interp_fraction=10,
        max_gamma=max_gamma,
        local_gamma=True,
        skip_once_passed=True,
        random_subset=random_subset,
    )

    return gamma


def local_gamma(
    filepath_ref,
    filepath_eval,
    result,
    random_subset=None,
    max_gamma=1.1,
    dose_threshold=1,
    distance_threshold=1,
):

    gamma = run_gamma(
        filepath_ref,
        filepath_eval,
        random_subset,
        max_gamma,
        dose_threshold,
        distance_threshold,
    )

    gamma_pass = calculate_pass_rate(gamma)

    assert np.round(gamma_pass, decimals=1) == result


@pytest.mark.pydicom
def test_max_gamma():
    local_gamma(
        get_data_file("H&N_VMAT_Reference_1mmPx.dcm"),
        get_data_file("H&N_VMAT_Evaluated_1mmPx.dcm"),
        93.6,
        random_subset=RANDOM_SUBSET,
        max_gamma=1.4,
    )

    local_gamma(
        get_data_file("H&N_VMAT_Reference_1mmPx.dcm"),
        get_data_file("H&N_VMAT_Evaluated_1mmPx.dcm"),
        93.6,
        random_subset=RANDOM_SUBSET,
        max_gamma=1.0001,
    )


@pytest.mark.pydicom
def test_local_gamma_1mm():
    local_gamma(
        get_data_file("H&N_VMAT_Reference_1mmPx.dcm"),
        get_data_file("H&N_VMAT_Evaluated_1mmPx.dcm"),
        93.6,
        random_subset=RANDOM_SUBSET,
    )


LOCAL_GAMMA_0_25_BASELINE = 96.9


@pytest.mark.pydicom
def test_local_gamma_0_25mm():
    local_gamma(
        get_data_file("H&N_VMAT_Reference_0_25mmPx.dcm"),
        get_data_file("H&N_VMAT_Evaluated_0_25mmPx.dcm"),
        LOCAL_GAMMA_0_25_BASELINE,
        random_subset=RANDOM_SUBSET,
    )


@pytest.mark.slow
@pytest.mark.pydicom
def test_multi_inputs():
    gamma = run_gamma(
        get_data_file("H&N_VMAT_Reference_0_25mmPx.dcm"),
        get_data_file("H&N_VMAT_Evaluated_0_25mmPx.dcm"),
        random_subset=RANDOM_SUBSET,
        max_gamma=1.0001,
        dose_threshold=[1, 0.2],
        distance_threshold=[1, 4],
    )

    baseline = {
        (1, 1): LOCAL_GAMMA_0_25_BASELINE,
        (1, 4): 99.8,  # 99.9 with higher interp_fraction
        (0.2, 1): 91.8,  # 94.4 with higher interp_fraction
        (0.2, 4): 99.2,  # 99.7 with higher interp_fraction
    }

    for key, value in gamma.items():
        assert np.round(calculate_pass_rate(value), decimals=1) == baseline[key]



# Copyright (C) 2019 Cancer Care Associates

# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at

#     http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


"""A command line interface for the conversion of Elekta binary log files.
"""


from pymedphys._trf.decode.detect import detect_cli
from pymedphys._trf.decode.trf2csv import trf2csv_cli
from pymedphys._trf.manage.orchestration import orchestration_cli


def trf_cli(subparsers):
    trf_parser = subparsers.add_parser(
        "trf",
        help=("A toolbox to work with the Elekta Linac ``.trf`` binary log files."),
    )
    trf_subparsers = trf_parser.add_subparsers(dest="trf")
    trf_to_csv(trf_subparsers)
    trf_detect(trf_subparsers)
    trf_orchestration(trf_subparsers)

    return trf_parser


def trf_to_csv(trf_subparsers):
    parser = trf_subparsers.add_parser(
        "to-csv", help="Converts ``.trf`` files to ``.csv`` table and header files."
    )

    parser.add_argument(
        "filepaths",
        type=str,
        nargs="+",
        help=(
            "A list of ``.trf`` filepaths that you wish to convert to ``.csv``. "
            "Use of the glob wildcard * is enabled, which means that running "
            "``pymedphys trf to-csv *.trf`` will convert all logfiles in the "
            "current directory to csv files."
        ),
    )

    parser.set_defaults(func=trf2csv_cli)


def trf_detect(trf_subparsers):
    parser = trf_subparsers.add_parser(
        "detect", help="Attempts to detect trf encoding method."
    )

    parser.add_argument("filepath", type=str, help=("The filepath of a trf file."))

    parser.set_defaults(func=detect_cli)


def trf_orchestration(trf_subparsers):
    parser = trf_subparsers.add_parser(
        "orchestrate",
        help=(
            """A command line interface for the management of trf files.

Requires two configuration csv files detailed as following:

``config_mosaiq_sql.csv``
-------------------------

::

    ,Timezone,Mosaiq SQL Server (Hostname:Port)
    a_centre,Australia/Sydney,mosaiq:1433
    another_centre,Australia/Sydney,another-mosaiq:1433


``config_linac_details.csv``
----------------------------

::

    ,Centre,IP
    1234,a_centre,192.168.150.40
    1236,a_centre,192.168.150.41
    1238,another_centre,10.0.0.40
"""
        ),
    )

    parser.add_argument(
        "data_directory", type=str, help="The path for storing the indexed log files."
    )

    parser.add_argument(
        "-m",
        "--mosaiq_sql",
        type=str,
        default=None,
        help=(
            "Define a custom path for the Mosaiq SQL configuration file. "
            "Defaults to ``{data_directory}/config_mosaiq_sql.csv``"
        ),
    )
    parser.add_argument(
        "-l",
        "--linac_details",
        type=str,
        default=None,
        help=(
            "Define a custom path for the Linac configuration file. "
            "Defaults to ``{data_directory}/config_linac_details.csv``"
        ),
    )

    parser.set_defaults(func=orchestration_cli)



# Copyright (C) 2018 Paul King

# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at

#     http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import numpy as np

from pymedphys._experimental.paulking.collequivalent import (
    mlc_equivalent_square_fs as equivalent_square,
)
from pymedphys._utilities.constants import A_LEAF_TYPE, AGILITY

# import pytest


def test_equivalent_mlc():
    """Compare effective field size for known pattern against benchmark."""
    mlc_segments = [(0.0, 0.0)] * 14
    mlc_segments += [
        (3.03, 2.47),
        (2.88, 2.46),
        (3.08, 2.51),
        (2.86, 2.46),
        (2.88, 2.46),
        (2.91, 5.04),
        (2.5, 5.04),
        (2.55, 4.87),
        (2.38, 4.61),
        (2.38, 7.04),
        (2.61, 7.46),
        (2.48, 6.55),
        (3.02, 6.52),
        (3.9, 7.2),
        (4.5, 7.5),
        (4.5, 7.5),
        (4.5, 7.5),
        (4.5, 7.5),
        (4.45, 7.5),
        (4.0, 7.5),
        (3.5, 7.5),
        (3.49, 7.5),
        (3.0, 7.5),
        (3.0, 7.5),
        (3.0, 7.5),
        (2.5, 7.5),
        (2.5, 7.5),
        (2.49, 6.52),
    ]
    mlc_segments += [(0.0, 0.0)] * 18

    mlc_segments = np.array(mlc_segments) * 10  # convert to mm

    assert abs(equivalent_square(mlc_segments, A_LEAF_TYPE) - 107.25) < 0.05


def an_equivalent_square(square_size):
    open_leaves = square_size // 5

    # Make sure evenly divides
    assert open_leaves == square_size / 5

    num_remaining_leaves = 80 - open_leaves
    leaves_on_top = num_remaining_leaves // 2
    leaves_on_bottom = num_remaining_leaves - leaves_on_top

    mlc_segments = (
        [(0, 0)] * leaves_on_top
        + [(square_size / 2, square_size / 2)] * open_leaves
        + [(0, 0)] * leaves_on_bottom
    )

    assert equivalent_square(mlc_segments, AGILITY) == square_size


# @pytest.mark.xfail
def test_equivalent_squares():
    sizes_to_test = (10, 20, 50, 100, 200, 400)

    for square_size in sizes_to_test:
        an_equivalent_square(square_size)



# Copyright (C) 2018 Simon Biggs

# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at

#     http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


import lzma
from glob import glob


def compress_test_file(filepath):
    with open(filepath, "rb") as load_file:
        with lzma.open("{}.xz".format(filepath), "w") as save_file:
            save_file.write(load_file.read())


def compress_test_files(glob_string, exclude_xz_files=True):
    files_to_compress = glob(glob_string, recursive=True)

    for filepath in files_to_compress:
        if not filepath.endswith(".xz") or not exclude_xz_files:
            compress_test_file(filepath)


def decompress_test_files():
    pass
